"""
Model class that holds the default project metrics data.

NOTE: If a user wants to customize the default project statistics screen by
      implementing their own statistics generator, they will likely need to
      create a custom version of this model class with their own set of metric
      fields. This class serves as an example of how project metrics and
      statistics can be structured and stored.

@relation(SDOC-SRS-97, scope=file)
"""

from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional

from strictdoc.backend.sdoc.models.node import SDocNode


@dataclass
class DocumentStats:
    requirements_total: int = 0
    requirements_no_uid: List[SDocNode] = field(default_factory=list)


@dataclass
class DocumentTreeStats:
    total_documents: int = 0
    total_requirements: int = 0
    total_sections: int = 0
    total_source_files: int = 0
    total_source_files_complete_coverage: int = 0
    total_source_files_partial_coverage: int = 0
    total_source_files_no_coverage: int = 0

    total_tbd: int = 0
    total_tbc: int = 0

    git_commit_hash: Optional[str] = None

    # Section.
    sections_without_text_nodes: int = 0

    # UID.
    requirements_no_uid: int = 0
    requirements_no_links: int = 0
    requirements_root_no_links: int = 0
    requirements_no_rationale: int = 0

    # STATUS.
    requirements_status_breakdown: Dict[Optional[str], int] = field(
        default_factory=lambda: defaultdict(int)
    )

    def sort_requirements_status_breakdown(self) -> None:
        self.requirements_status_breakdown = dict(
            sorted(
                self.requirements_status_breakdown.items(),
                key=lambda item: (item[0] is not None, item[1]),
                reverse=True,
            )
        )



"""
@relation(SDOC-SRS-98, SDOC-SRS-109, scope=file)
"""

from collections import defaultdict
from dataclasses import dataclass
from typing import DefaultDict, Dict, Generator, List, Optional, Set

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document_config import DocumentConfig
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.document_view import DocumentView
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementField,
    GrammarElementFieldMultipleChoice,
    GrammarElementFieldSingleChoice,
    GrammarElementFieldTag,
)
from strictdoc.backend.sdoc.models.model import (
    SDocDocumentFromFileIF,
    SDocDocumentIF,
    SDocElementIF,
    SDocNodeIF,
)
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.core.document_meta import DocumentMeta
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.mid import MID
from strictdoc.helpers.ordered_set import OrderedSet
from strictdoc.helpers.string import tokenize


@auto_described
class SDocDocumentContext:
    def __init__(self) -> None:
        self.title_number_string: Optional[str] = None


@dataclass
class SDocDocumentSearchIndex:
    document_index: DefaultDict[str, Set[str]]
    map_nodes_by_mid: Dict[str, Dict[str, str]]

    @classmethod
    def create_empty(cls) -> "SDocDocumentSearchIndex":
        return SDocDocumentSearchIndex(
            document_index=defaultdict(set), map_nodes_by_mid={}
        )


@auto_described
class SDocDocument(SDocDocumentIF):
    def __init__(
        self,
        mid: Optional[str],
        title: str,
        config: Optional[DocumentConfig],
        view: Optional[DocumentView],
        grammar: Optional[DocumentGrammar],
        section_contents: List[SDocElementIF],
        is_bundle_document: bool = False,
    ) -> None:
        self.title: str = title
        self.reserved_title: str = title
        self.config: DocumentConfig = (
            config
            if config is not None
            else DocumentConfig.default_config(self)
        )
        self.view: DocumentView = (
            view if view is not None else DocumentView.create_default(self)
        )
        self.grammar: Optional[DocumentGrammar] = grammar
        self.section_contents: List[SDocElementIF] = section_contents

        self.is_bundle_document: bool = is_bundle_document

        self.fragments_from_files: List[SDocDocumentFromFileIF] = []

        # FIXME: Is this used?
        self.ng_level: int = 0
        self.ng_has_requirements = False

        self.meta: Optional[DocumentMeta] = None

        self.reserved_mid: MID = MID(mid) if mid is not None else MID.create()
        self.mid_permanent: bool = mid is not None
        self.included_documents: List[SDocDocumentIF] = []
        self.context: SDocDocumentContext = SDocDocumentContext()

        self.ng_including_document_reference: Optional[DocumentReference] = None
        self.ng_including_document_from_file: Optional[
            SDocDocumentFromFileIF
        ] = None

        self.ng_whitelisted: bool = True
        self.search_index = SDocDocumentSearchIndex.create_empty()

    def iterate_nodes(
        self, element_type: Optional[str] = None
    ) -> Generator[SDocNodeIF, None, None]:
        """
        Iterate over all non-[TEXT] nodes in the document.

        If element_type is given, then only nodes of type `element_type` are
        returned. Otherwise, all element types are returned.
        """
        task_list: List[SDocElementIF] = list(self.section_contents)
        while task_list:
            node = task_list.pop(0)

            if isinstance(node, SDocDocumentFromFileIF):
                yield from node.iterate_nodes(element_type)

            if isinstance(node, SDocNodeIF):
                if node.node_type != "TEXT":
                    if element_type is None or node.node_type == element_type:
                        yield node

            task_list.extend(node.section_contents)

    def has_any_requirements(self) -> bool:
        return any(True for _ in self.iterate_nodes())

    def collect_options_for_tag(
        self, element_type: str, field_name: str
    ) -> List[str]:
        """
        Returns the list of existing options for a tag field in this document.
        """
        option_set: OrderedSet[str] = OrderedSet()

        for nodeif in self.iterate_nodes(element_type):
            node = assert_cast(nodeif, SDocNode)
            if field_name in node.ordered_fields_lookup:
                node_field = node.ordered_fields_lookup[field_name][0]
                field_value = node_field.get_text_value()
                if field_value:
                    options = [
                        option.strip()
                        for option in field_value.split(",")
                        if option.strip()
                    ]
                    for option in options:
                        option_set.add(option)

        return list(option_set)

    @property
    def uid(self) -> Optional[str]:
        return self.config.uid

    def is_section(self) -> bool:
        return True

    @property
    def is_root_included_document(self) -> bool:
        return self.document_is_included()

    def is_requirement(self) -> bool:
        return False

    def get_display_node_type(self) -> str:
        return "Document"

    def get_node_type_string(self) -> Optional[str]:
        return None

    def get_type_string(self) -> str:
        return "document" if not self.document_is_included() else "section"

    def get_debug_info(self) -> str:
        debug_components: List[str] = [f"TITLE = '{self.title}'"]
        if self.meta is not None:
            debug_components.append(
                f" ({self.meta.input_doc_rel_path.relative_path})"
            )
        return f"Document({', '.join(debug_components)})"

    def document_is_included(self) -> bool:
        if self.ng_including_document_reference is None:
            return False
        return self.ng_including_document_reference.get_document() is not None

    def get_including_document(self) -> Optional["SDocDocumentIF"]:
        if self.ng_including_document_reference is None:
            return None
        return self.ng_including_document_reference.get_document()

    def iterate_included_documents_depth_first(
        self,
    ) -> Generator["SDocDocumentIF", None, None]:
        for included_document_ in self.included_documents:
            yield included_document_
            yield from included_document_.iterate_included_documents_depth_first()

    @property
    def reserved_uid(self) -> Optional[str]:
        return self.config.uid

    def assign_meta(self, meta: DocumentMeta) -> None:
        assert isinstance(meta, DocumentMeta)
        self.meta = meta

    def has_any_nodes(self) -> bool:
        return len(self.section_contents) > 0

    def get_display_title(
        self,
        include_toc_number: bool = True,  # noqa: ARG002
    ) -> str:
        return self.title

    @property
    def ng_resolved_custom_level(self) -> Optional[str]:
        return None

    @property
    def requirement_prefix(self) -> str:
        return self.get_prefix()

    def get_prefix(self) -> str:
        return self.config.get_prefix()

    def get_prefix_for_new_node(self, node_type: str) -> Optional[str]:
        assert isinstance(node_type, str) and len(node_type), node_type

        grammar: DocumentGrammar = assert_cast(self.grammar, DocumentGrammar)
        element: GrammarElement = grammar.elements_by_type[node_type]
        if (element_prefix := element.property_prefix) is not None:
            if element_prefix == "None":
                return None
            return element_prefix

        return self.get_prefix()

    def enumerate_meta_field_titles(self) -> Generator[str, None, None]:
        assert self.grammar is not None
        assert self.grammar.elements is not None
        # FIXME: currently only enumerating a single element ([0])
        yield from self.grammar.elements[0].enumerate_meta_field_titles()

    def enumerate_custom_content_field_titles(
        self,
    ) -> Generator[str, None, None]:
        assert self.grammar is not None
        assert self.grammar.elements is not None
        # FIXME: currently only enumerating a single element ([0])
        yield from self.grammar.elements[
            0
        ].enumerate_custom_content_field_titles()

    def get_grammar_element_field_for(
        self, element_type: str, field_name: str
    ) -> GrammarElementField:
        """
        Returns the GrammarElementField for a field of a [element_type] in this document.
        """
        grammar: DocumentGrammar = assert_cast(self.grammar, DocumentGrammar)
        element: GrammarElement = grammar.elements_by_type[element_type]
        field: GrammarElementField = element.fields_map[field_name]
        return field

    def get_options_for_field(
        self, element_type: str, field_name: str
    ) -> List[str]:
        """
        Returns the list of valid options for a Single/MultiChoice field in this document.
        """
        field: GrammarElementField = self.get_grammar_element_field_for(
            element_type, field_name
        )

        if isinstance(field, GrammarElementFieldSingleChoice) or isinstance(
            field, GrammarElementFieldMultipleChoice
        ):
            return field.options

        if isinstance(field, GrammarElementFieldTag):
            return self.collect_options_for_tag(element_type, field_name)

        raise AssertionError(f"Must not reach here: {field}")

    def blacklist_if_needed(self) -> None:
        if self.section_contents is not None:
            for node_ in self.section_contents:
                if node_.ng_whitelisted:
                    return

        self.ng_whitelisted = False

    def build_search_index(self) -> None:
        """
        Build a static search index for this document.

        @relation(SDOC-SRS-155, scope=function)
        """

        document_index = defaultdict(set)
        map_nodes_by_mid = {}

        from strictdoc.core.document_iterator import (  # noqa: PLC0415
            DocumentCachingIterator,
        )

        document_iterator = DocumentCachingIterator(self)

        for node, _ in document_iterator.all_content(
            print_fragments=False,
        ):
            if not isinstance(node, SDocNode):
                continue

            node_dict = {}

            node_dict["MID"] = node.reserved_mid.get_string_value()
            map_nodes_by_mid[node.reserved_mid.get_string_value()] = node_dict

            for (
                field_name_,
                field_values_,
            ) in node.ordered_fields_lookup.items():
                requirement_field: SDocNodeField = field_values_[0]
                requirement_field_value = requirement_field.get_text_value()

                node_dict[field_name_] = requirement_field_value

                tokens = set(tokenize(requirement_field_value))
                for token in tokens:
                    if len(token) > 1:
                        document_index[token].add(
                            node.reserved_mid.get_string_value()
                        )

                        for i in range(0, len(token)):
                            token_incremental = token[: i + 1]
                            document_index[token_incremental].add(
                                node.reserved_mid
                            )
                            token_deincremental = token[i:]
                            document_index[token_deincremental].add(
                                node.reserved_mid
                            )

        self.search_index = SDocDocumentSearchIndex(
            document_index, map_nodes_by_mid
        )



"""
@relation(SDOC-SRS-95, scope=file)
"""

import hashlib
import os
from pathlib import Path
from typing import Any

from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.md5 import get_file_md5
from strictdoc.helpers.pickle import pickle_dump, pickle_load


class PickleCache:
    @staticmethod
    def read_from_cache(
        file_path: str, project_config: ProjectConfig, content_kind: str
    ) -> Any:
        path_to_cached_file: str = PickleCache.get_cached_file_path(
            file_path, project_config, content_kind
        )
        if os.path.isfile(path_to_cached_file):
            with open(path_to_cached_file, "rb") as cache_file:
                unpickled_content = cache_file.read()
            try:
                return pickle_load(unpickled_content)
            except Exception as exception_:
                raise AssertionError(
                    "MUST NOT REACH HERE: "
                    f"Error when unpickling a cache file: {path_to_cached_file}. "
                    "To fix the issue, simply remove the cache file or the whole cache folder. "
                    "Please report this exception to StrictDoc developers: "
                    f"https://github.com/strictdoc-project/strictdoc/issues/new"
                ) from exception_
        return None

    @staticmethod
    def save_to_cache(
        content: Any,
        file_path: str,
        project_config: ProjectConfig,
        content_kind: str,
    ) -> None:
        path_to_cached_file: str = PickleCache.get_cached_file_path(
            file_path, project_config, content_kind
        )
        path_to_cached_file_dir: str = os.path.dirname(path_to_cached_file)
        Path(path_to_cached_file_dir).mkdir(parents=True, exist_ok=True)
        pickled_content: Any = pickle_dump(content)
        with open(path_to_cached_file, "wb") as cache_file:
            cache_file.write(pickled_content)

    @staticmethod
    def get_cached_file_path(
        file_path: str, project_config: ProjectConfig, content_kind: str
    ) -> str:
        path_to_tmp_dir = project_config.get_path_to_cache_dir()

        full_path_to_file = (
            file_path
            if os.path.isabs(file_path)
            else os.path.abspath(file_path)
        )

        file_md5: str = get_file_md5(file_path)

        # File name contains an MD5 hash of its full path to ensure the
        # uniqueness of the cached items. Additionally, the unique file name
        # contains a full path to the output root to prevent collisions
        # between StrictDoc invocations running against the same set of SDoc
        # files in parallel.
        unique_identifier = project_config.output_dir + full_path_to_file
        unique_identifier_md5 = hashlib.md5(
            unique_identifier.encode("utf-8")
        ).hexdigest()
        file_name = os.path.basename(full_path_to_file)
        file_name += "_" + unique_identifier_md5 + "_" + file_md5

        path_to_cached_file = os.path.join(
            path_to_tmp_dir,
            content_kind,
            file_name,
        )

        return path_to_cached_file



import os
from pathlib import Path

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.core.document_meta import DocumentMeta
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.rst.writer import RSTWriter
from strictdoc.helpers.cast import assert_cast


class DocumentRSTGenerator:
    @staticmethod
    def export_tree(
        traceability_index: TraceabilityIndex, output_rst_root: str
    ) -> None:
        Path(output_rst_root).mkdir(parents=True, exist_ok=True)

        document: SDocDocument
        for document in traceability_index.document_tree.document_list:
            document_content = DocumentRSTGenerator.export(
                document, traceability_index
            )
            document_meta = assert_cast(document.meta, DocumentMeta)
            output_folder = os.path.join(
                output_rst_root,
                document_meta.input_doc_dir_rel_path.relative_path,
            )
            Path(output_folder).mkdir(parents=True, exist_ok=True)

            document_out_file = f"{document_meta.document_filename_base}.rst"
            output_path = os.path.join(output_folder, document_out_file)

            with open(output_path, "w", encoding="utf8") as file:
                file.write(document_content)

    @staticmethod
    def export(
        document: SDocDocument, traceability_index: TraceabilityIndex
    ) -> str:
        writer = RSTWriter(traceability_index)

        single_or_many = (
            len(traceability_index.document_tree.document_list) == 1
        )
        output = writer.write(document, single_or_many)

        return output



import os


def path_to_posix_path(path: str) -> str:
    return path.replace("\\", "/")


class SDocRelativePath:
    """
    Class to encapsulate a path as both a native OS path and a POSIX path.

    A native OS path can be with forward slashes on Linux and macOS and with
    backward slashes on Windows. The POSIX path is always with forward slashes.

    This rel path will be used for lookups of URL pages that are
    always "/"-based, even on Windows.
    """

    def __init__(self, relative_path: str) -> None:
        assert isinstance(relative_path, str), relative_path
        self.relative_path: str = relative_path
        self.relative_path_posix: str = path_to_posix_path(relative_path)

    @staticmethod
    def from_url(url: str) -> "SDocRelativePath":
        return SDocRelativePath(os.path.normpath(url).replace("/", os.sep))

    def length(self) -> int:
        return len(self.relative_path)

    def __str__(self) -> str:
        raise AssertionError("Must not be used")

    def __repr__(self) -> str:
        raise AssertionError("Must not be used")



import os
from copy import deepcopy
from datetime import datetime

from strictdoc.cli.cli_arg_parser import DiffCommandConfig
from strictdoc.core.project_config import ProjectConfig
from strictdoc.export.html.generators.view_objects.diff_screen_results_view_object import (
    DiffScreenResultsViewObject,
)
from strictdoc.export.html.html_generator import HTMLGenerator
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.git.change_generator import ChangeContainer, ChangeGenerator


class DiffCommand:
    @staticmethod
    def execute(
        *, project_config: ProjectConfig, diff_config: DiffCommandConfig
    ) -> None:
        if not project_config.is_activated_diff():
            raise RuntimeError(
                "The DIFF feature is not activated in the project config."
            )

        # FIXME: Handle merging of Diff config data into project config more
        # gracefully.
        if diff_config.output_dir is not None:
            assert os.path.isdir(diff_config.output_dir)
            project_config.export_output_html_root = diff_config.output_dir
        else:
            project_config.export_output_html_root = os.path.join(
                os.getcwd(), "output"
            )

        assert os.path.isdir(diff_config.path_to_lhs_tree)
        assert os.path.isdir(diff_config.path_to_rhs_tree)

        if os.path.isabs(diff_config.path_to_lhs_tree):
            lhs_export_input_abs_path = diff_config.path_to_lhs_tree
        else:
            lhs_export_input_abs_path = os.path.join(
                os.getcwd(), diff_config.path_to_lhs_tree
            )
        if os.path.isabs(diff_config.path_to_rhs_tree):
            rhs_export_input_abs_path = diff_config.path_to_rhs_tree
        else:
            rhs_export_input_abs_path = os.path.join(
                os.getcwd(), diff_config.path_to_rhs_tree
            )

        project_config_copy_lhs: ProjectConfig = deepcopy(project_config)
        project_config_copy_rhs: ProjectConfig = deepcopy(project_config)

        project_config_copy_lhs.input_paths = [lhs_export_input_abs_path]
        project_config_copy_rhs.input_paths = [rhs_export_input_abs_path]

        change_container: ChangeContainer = ChangeGenerator.generate(
            lhs_project_config=project_config_copy_lhs,
            rhs_project_config=project_config_copy_rhs,
        )

        html_templates = HTMLTemplates.create(
            project_config=project_config,
            enable_caching=False,
            strictdoc_last_update=datetime.today(),
        )

        assert change_container.traceability_index_lhs.document_tree is not None
        assert change_container.traceability_index_rhs.document_tree is not None

        view_object = DiffScreenResultsViewObject(
            project_config=project_config,
            change_container=change_container,
            document_tree_lhs=change_container.traceability_index_lhs.document_tree,
            document_tree_rhs=change_container.traceability_index_rhs.document_tree,
            documents_iterator_lhs=change_container.documents_iterator_lhs,
            documents_iterator_rhs=change_container.documents_iterator_rhs,
            left_revision=lhs_export_input_abs_path,
            right_revision=rhs_export_input_abs_path,
            lhs_stats=change_container.lhs_stats,
            rhs_stats=change_container.rhs_stats,
            change_stats=change_container.change_stats,
            traceability_index_lhs=change_container.traceability_index_lhs,
            traceability_index_rhs=change_container.traceability_index_rhs,
            tab="diff",
        )

        output = view_object.render_screen(html_templates.jinja_environment())
        path_to_output_file = os.path.join(
            project_config.export_output_html_root, "diff.html"
        )
        with open(path_to_output_file, "w", encoding="utf8") as output_file:
            output_file.write(output)

        # Diff summary generator...

        view_object = DiffScreenResultsViewObject(
            project_config=project_config,
            change_container=change_container,
            document_tree_lhs=change_container.traceability_index_lhs.document_tree,
            document_tree_rhs=change_container.traceability_index_rhs.document_tree,
            documents_iterator_lhs=change_container.documents_iterator_lhs,
            documents_iterator_rhs=change_container.documents_iterator_rhs,
            left_revision=lhs_export_input_abs_path,
            right_revision=rhs_export_input_abs_path,
            lhs_stats=change_container.lhs_stats,
            rhs_stats=change_container.rhs_stats,
            change_stats=change_container.change_stats,
            traceability_index_lhs=change_container.traceability_index_lhs,
            traceability_index_rhs=change_container.traceability_index_rhs,
            tab="changelog",
        )
        output = view_object.render_screen(html_templates.jinja_environment())
        path_to_output_file = os.path.join(
            project_config.export_output_html_root, "changelog.html"
        )
        with open(path_to_output_file, "w", encoding="utf8") as output_file:
            output_file.write(output)

        # Copy assets.
        html_generator = HTMLGenerator(
            project_config=project_config, html_templates=html_templates
        )
        html_generator.export_assets(
            traceability_index=change_container.traceability_index_lhs,
            project_config=project_config,
            export_output_html_root=project_config.export_output_html_root,
        )



QUERY_GRAMMAR = r"""
Query:
 root_expression = BooleanExpression
;

AndExpression:
  '('
  expressions += BooleanExpression
  ('and' expressions += BooleanExpression)+
  ')'
;

OrExpression:
  '('
  expressions += BooleanExpression
  ('or' expressions += BooleanExpression)+
  ')'
;

NotExpression:
  'not'
  expression = BooleanExpression
;

BooleanExpression:
  AndExpression
  |
  OrExpression
  |
  NotExpression
  |
  NodeHasParentRequirementsExpression
  |
  NodeHasChildRequirementsExpression
  |
  InExpression
  |
  NotInExpression
  |
  NodeContainsExpression
  |
  NodeContainsAnyFreeTextExpression
  |
  NodeIsRequirementExpression
  |
  NodeIsSectionExpression
  |
  NodeIsSourceFileExpression
  |
  NodeIsSourceFileWithCompleteCoverageExpression
  |
  NodeIsSourceFileWithPartialCoverageExpression
  |
  NodeIsSourceFileWithNoCoverageExpression
  |
  NodeIsRootExpression
  |
  EqualExpression
  |
  NotEqualExpression
;

StringExpression:
  '"' string = /[^"]+/ '"'
;

NoneExpression:
  _ = 'None'
;

NodeFieldExpression:
  'node["' field_name = /[A-Za-z0-9]+/ '"]'
;

NodeContainsExpression:
  'node.contains("' string = /[A-Za-z0-9]+/ '")'
;

NodeContainsAnyFreeTextExpression:
  _ = 'node.contains_any_text'
;

NodeHasParentRequirementsExpression:
  _ = 'node.has_parent_requirements'
;

NodeHasChildRequirementsExpression:
  _ = 'node.has_child_requirements'
;

NodeIsRequirementExpression:
  _ = 'node.is_requirement' '()'?
;

NodeIsRootExpression:
  _ = 'node.is_root'
;

NodeIsSectionExpression:
  _ = 'node.is_section' '()'?
;

NodeIsSourceFileExpression:
  _ = 'node.is_source_file()'
;

NodeIsSourceFileWithCompleteCoverageExpression:
  _ = 'node.is_source_file_with_complete_coverage()'
;

NodeIsSourceFileWithPartialCoverageExpression:
  _ = 'node.is_source_file_with_partial_coverage()'
;

NodeIsSourceFileWithNoCoverageExpression:
  _ = 'node.is_source_file_with_no_coverage()'
;

EqualExpression:
  lhs_expr = ComparableExpression '==' rhs_expr = ComparableExpression
;

NotEqualExpression:
  lhs_expr = ComparableExpression '!=' rhs_expr = ComparableExpression
;

ComparableExpression:
  NodeFieldExpression | StringExpression | NoneExpression
;

InExpression:
  lhs_expr = InableLHSExpression 'in' rhs_expr = InableRHSExpression
;

NotInExpression:
  lhs_expr = InableLHSExpression 'not in' rhs_expr = InableRHSExpression
;

InableLHSExpression:
  NodeFieldExpression | StringExpression
;

InableRHSExpression:
  NodeFieldExpression | StringExpression
;

"""



import hashlib


def get_md5(obj: str) -> str:
    return hashlib.md5(obj.encode("utf-8")).hexdigest()


def get_file_md5(path: str, buf_size: int = 65536) -> str:
    m = hashlib.md5()
    with open(path, "rb") as f:
        b = f.read(buf_size)
        while len(b) > 0:
            m.update(b)
            b = f.read(buf_size)
    return m.hexdigest()



from collections import OrderedDict
from typing import Dict, List, Optional, Set, Tuple

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementField,
)
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.export.html.form_objects.grammar_element_form_object import (
    GrammarElementFormObject,
)
from strictdoc.helpers.cast import assert_cast


class UpdateGrammarElementCommand:
    def __init__(
        self,
        *,
        form_object: GrammarElementFormObject,
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
    ):
        self.form_object: GrammarElementFormObject = form_object
        self.document: SDocDocument = document
        self.traceability_index: TraceabilityIndex = traceability_index

    def perform(self) -> None:
        form_object: GrammarElementFormObject = self.form_object
        document: SDocDocument = self.document
        assert document.grammar is not None

        existing_element: GrammarElement = document.grammar.get_element_by_mid(
            form_object.element_mid
        )

        grammar_fields: Dict[str, GrammarElementField] = {}
        for grammar_field in existing_element.fields:
            grammar_fields[grammar_field.mid] = grammar_field

        # Prepare fields that could have been renamed by the user has just saved the form.
        renamed_fields_lookup = {}
        for field in form_object.fields:
            if field.field_mid not in grammar_fields:
                continue
            existing_field = grammar_fields[field.field_mid]
            if field.field_name != existing_field.title:
                renamed_fields_lookup[field.field_name] = existing_field.title

        #
        # Convert the form object to an updated grammar element.
        #
        assert document.grammar is not None
        updated_element: GrammarElement = (
            form_object.convert_to_grammar_element(document.grammar)
        )

        #
        # Compare if anything was changed in the new grammar.
        #
        document_grammar_field_names = updated_element.get_field_titles()

        existing_requirement_element = document.grammar.elements_by_type[
            existing_element.tag
        ]

        document.grammar.update_element(existing_element, updated_element)

        document_iterator = self.traceability_index.document_iterators[document]

        for node, _ in document_iterator.all_content():
            if not isinstance(node, SDocNode):
                continue

            requirement: SDocNode = assert_cast(node, SDocNode)
            if requirement.node_type != updated_element.tag:
                continue

            requirement_field_names = list(
                requirement.ordered_fields_lookup.keys()
            )

            # Rewrite requirement fields because some fields could have been
            # renamed.
            new_ordered_fields_lookup: OrderedDict[str, List[SDocNodeField]] = (
                OrderedDict()
            )

            for document_grammar_field_name in document_grammar_field_names:
                # We need to find a previous field name in case the field was
                # renamed.
                previous_field_name = renamed_fields_lookup.get(
                    document_grammar_field_name, document_grammar_field_name
                )

                # If the field does not exist in the grammar fields anymore,
                # delete the requirement field.
                if previous_field_name not in requirement_field_names:
                    continue

                previous_fields: List[SDocNodeField] = (
                    requirement.ordered_fields_lookup[previous_field_name]
                )
                for previous_field in previous_fields:
                    previous_field.field_name = document_grammar_field_name

                new_ordered_fields_lookup[document_grammar_field_name] = (
                    previous_fields
                )

            registered_relation_types: Set[Tuple[str, Optional[str]]] = set()
            for relation in existing_requirement_element.relations:
                registered_relation_types.add(
                    (relation.relation_type, relation.relation_role)
                )
            if len(requirement.relations) > 0:
                new_relations = []
                for requirement_relation_ in requirement.relations:
                    if (
                        requirement_relation_.ref_type,
                        requirement_relation_.role,
                    ) in registered_relation_types:
                        new_relations.append(requirement_relation_)
                requirement.relations = new_relations
            requirement.ordered_fields_lookup = new_ordered_fields_lookup



import os
from typing import Optional, Union

import bs4
from bs4 import BeautifulSoup
from datauri import DataURI


class EmbeddableTag:
    ICO = "ico"
    CSS = "css"
    JS = "js"
    IMAGE_PNG = "image_png"
    IMAGE_SVG = "image_svg"

    MAP = {
        CSS: {"attr": "href", "type": "text/css"},
        JS: {"attr": "src", "type": "text/javascript"},
        ICO: {"attr": "href", "type": "image/x-icon"},
        IMAGE_PNG: {"attr": "src", "type": "image/png"},
        IMAGE_SVG: {"attr": "data", "type": "image/svg+xml"},
    }

    def __init__(self, asset_type: str, path: str) -> None:
        assert asset_type
        assert path
        self.asset_type: str = asset_type
        self.path: str = path

    @staticmethod
    def recognize_from_soup_tag(
        tag: bs4.element.Tag,
    ) -> Optional["EmbeddableTag"]:
        rel_value: str
        if tag.name == "link":
            if "rel" in tag.attrs:
                if len(tag.attrs["rel"]) == 1:
                    rel_value = tag.attrs["rel"][0]
                    if rel_value == "stylesheet":
                        return EmbeddableTag(
                            EmbeddableTag.CSS, str(tag.attrs["href"])
                        )
                if len(tag.attrs["rel"]) == 2:
                    rel_value = tag.attrs["rel"][0]
                    if rel_value == "shortcut":
                        ref_value_second = tag.attrs["rel"][1]
                        if ref_value_second == "icon":
                            return EmbeddableTag(
                                EmbeddableTag.ICO, str(tag.attrs["href"])
                            )
        elif tag.name == "script":
            if "type" in tag.attrs:
                type_value = tag.attrs["type"]
                if (
                    type_value in ("text/javascript", "module")
                    and "src" in tag.attrs
                ):
                    return EmbeddableTag(
                        EmbeddableTag.JS, str(tag.attrs["src"])
                    )
        elif tag.name == "img":
            if "src" in tag.attrs:
                rel_value = str(tag.attrs["src"])
                if rel_value.lower().endswith(".png"):
                    return EmbeddableTag(
                        EmbeddableTag.IMAGE_PNG, str(tag.attrs["src"])
                    )
        elif tag.name == "object":
            if "type" in tag.attrs:
                type_value = tag.attrs["type"]
                if type_value == "image/svg+xml":
                    return EmbeddableTag(
                        EmbeddableTag.IMAGE_SVG, str(tag.attrs["data"])
                    )
        return None

    def get_path(self) -> str:
        return self.path

    def get_attr(self) -> str:
        return EmbeddableTag.MAP[self.asset_type]["attr"]

    def get_type(self) -> str:
        return EmbeddableTag.MAP[self.asset_type]["type"]


class HTMLEmbedder:
    @staticmethod
    def embed_assets(html_string: str, path: str) -> str:
        soup = BeautifulSoup(html_string, "html5lib")

        tag: Union[
            bs4.element.Tag,
            bs4.element.PageElement,
            bs4.element.NavigableString,
        ]
        for tag in soup.find_all(recursive=True):
            if not isinstance(tag, bs4.element.Tag):
                continue
            embeddable_tag = EmbeddableTag.recognize_from_soup_tag(tag)
            if not embeddable_tag:
                continue
            output_dir = os.path.dirname(path)
            asset_path = os.path.join(output_dir, embeddable_tag.get_path())
            tag.attrs[embeddable_tag.get_attr()] = (
                HTMLEmbedder._read_file_as_base64(asset_path)
            )

        output = str(soup)
        return output

    @staticmethod
    def _read_file_as_base64(asset_path: str) -> DataURI:
        # DataURI.from_file seems to work well without knowing content type.
        # There is also a lower-level DataURI.make file.
        base64 = DataURI.from_file(asset_path)
        return base64



# https://stackoverflow.com/questions/21778252/how-to-raise-an-exception-in-a-jinja2-macro
from typing import Any, List, Optional, Union

from jinja2 import Environment, TemplateRuntimeError, nodes
from jinja2.ext import Extension
from jinja2.parser import Parser


class AssertExtension(Extension):
    tags = {"assert"}

    def __init__(self, environment: Environment) -> None:
        super().__init__(environment)
        self.current_line: int = -1
        self.current_file: Optional[str] = None

    def parse(self, parser: Parser) -> Union[nodes.Node, List[nodes.Node]]:
        lineno = next(parser.stream).lineno
        self.current_line = lineno
        self.current_file = parser.filename

        condition_node = parser.parse_expression()
        if parser.stream.skip_if("comma"):
            context_node = parser.parse_expression()
        else:
            context_node = nodes.Const(None)

        return nodes.CallBlock(
            self.call_method(
                "_assert", [condition_node, context_node], lineno=lineno
            ),
            [],
            [],
            [],
            lineno=lineno,
        )

    def _assert(
        self,
        condition: bool,
        context_or_none: Optional[Any],
        caller: Any,  # noqa: ARG002
    ) -> str:
        if not condition:
            error_message = (
                f"Assertion error in the Jinja template: "
                f"{self.current_file}:{self.current_line}."
            )
            if context_or_none:
                error_message += f" Message: {context_or_none}"
            raise TemplateRuntimeError(error_message)
        return ""



# pragma: no cover file
from dataclasses import dataclass
from typing import Optional

from strictdoc import __version__
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_view import DocumentView
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.document_tree_iterator import DocumentTreeIterator
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.html_templates import HTMLTemplates, JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.helpers.cast import assert_cast


@dataclass
class NestorViewObject:
    def __init__(
        self,
        *,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
        templates: HTMLTemplates,
        path_to_json: str,
    ):
        self.traceability_index: TraceabilityIndex = traceability_index
        self.project_config: ProjectConfig = project_config
        self.templates: HTMLTemplates = templates
        self.path_to_json: str = path_to_json

        link_renderer = LinkRenderer(
            root_path="", static_path=project_config.dir_for_sdoc_assets
        )
        markup_renderer = MarkupRenderer.create(
            "RST",
            traceability_index,
            link_renderer,
            templates,
            project_config,
            None,
        )

        self.link_renderer: LinkRenderer = link_renderer
        self.markup_renderer: MarkupRenderer = markup_renderer
        self.standalone: bool = False
        self.document_tree_iterator: DocumentTreeIterator = (
            DocumentTreeIterator(
                assert_cast(traceability_index.document_tree, DocumentTree)
            )
        )
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__
        self.current_view = DocumentView.create_default(None).views[0]
        self.document_type: DocumentType = DocumentType.DOCUMENT
        self.link_document_type: DocumentType = DocumentType.DOCUMENT
        self.document: Optional[SDocDocument] = None

    def render_screen(self, jinja_environment: JinjaEnvironment) -> str:
        return jinja_environment.render_template_as_markup(
            "screens/nestor/index.jinja", view_object=self
        )

    def render_static_url(self, url: str) -> str:
        return self.link_renderer.render_static_url(url)



from typing import Any, Dict, List, Optional, Tuple

from docutils import nodes
from docutils.parsers.rst.states import Inliner


def raw_html_role(
    _name: str,
    _rawtext: str,
    text: str,
    _lineno: int,
    _inliner: Inliner,
    _options: Optional[Dict[str, Any]] = None,
    _content: Optional[List[str]] = None,
) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
    html = nodes.raw(text, text, format="html")
    return [html], []



from typing import Dict, Optional, Tuple, Union

from markupsafe import Markup

from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.model import RequirementFieldName
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html.renderers.html_fragment_writer import (
    HTMLFragmentWriter,
)
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.text_to_html_writer import TextToHtmlWriter
from strictdoc.export.rst.rst_to_html_fragment_writer import (
    RstToHtmlFragmentWriter,
)
from strictdoc.helpers.rst import escape_str_after_inline_markup


class MarkupRenderer:
    @staticmethod
    def create(
        markup: Optional[str],
        traceability_index: TraceabilityIndex,
        link_renderer: LinkRenderer,
        html_templates: HTMLTemplates,
        config: ProjectConfig,
        context_document: Optional[SDocDocument],
    ) -> "MarkupRenderer":
        assert isinstance(html_templates, HTMLTemplates)
        html_fragment_writer: Union[
            RstToHtmlFragmentWriter,
            HTMLFragmentWriter,
            TextToHtmlWriter,
        ]
        if not markup or markup == "RST":
            html_fragment_writer = RstToHtmlFragmentWriter(
                project_config=config,
                context_document=context_document,
            )
        elif markup == "HTML":
            html_fragment_writer = HTMLFragmentWriter()
        else:
            html_fragment_writer = TextToHtmlWriter()
        return MarkupRenderer(
            html_fragment_writer,
            traceability_index,
            link_renderer,
            html_templates,
            context_document,
        )

    def __init__(
        self,
        fragment_writer: Union[
            RstToHtmlFragmentWriter,
            HTMLFragmentWriter,
            TextToHtmlWriter,
        ],
        traceability_index: TraceabilityIndex,
        link_renderer: LinkRenderer,
        html_templates: HTMLTemplates,
        context_document: Optional[SDocDocument],
    ) -> None:
        assert isinstance(traceability_index, TraceabilityIndex)
        assert isinstance(link_renderer, LinkRenderer)
        assert context_document is None or isinstance(
            context_document, SDocDocument
        ), context_document
        assert isinstance(html_templates, HTMLTemplates)

        self.fragment_writer: Union[
            RstToHtmlFragmentWriter,
            HTMLFragmentWriter,
            TextToHtmlWriter,
        ] = fragment_writer
        self.traceability_index: TraceabilityIndex = traceability_index
        self.link_renderer: LinkRenderer = link_renderer
        self.context_document: Optional[SDocDocument] = context_document

        # FIXME: Now that the underlying RST fragment caching is in place,
        # This caching could be removed. It is unlikely that it adds any serious
        # performance improvement.
        self.cache: Dict[Tuple[DocumentType, SDocNodeField], Markup] = {}

        self.template_anchor = html_templates.jinja_environment().get_template(
            "rst/anchor.jinja"
        )

    def render_node_statement(
        self, document_type: DocumentType, node: SDocNode
    ) -> Markup:
        assert isinstance(node, SDocNode)
        return self.render_node_field(document_type, node.get_content_field())

    def render_truncated_node_statement(
        self, document_type: DocumentType, node: SDocNode
    ) -> Markup:
        assert isinstance(node, SDocNode)
        # FIXME: Double-check and switch to truncating using CSS.
        # https://github.com/strictdoc-project/strictdoc/issues/1925
        return self.render_node_field(document_type, node.get_content_field())

    def render_node_rationale(
        self, document_type: DocumentType, node: SDocNode
    ) -> Markup:
        assert isinstance(node, SDocNode)
        return self.render_node_field(
            document_type,
            node.get_field_by_name(RequirementFieldName.RATIONALE),
        )

    def render_node_field(
        self, document_type: DocumentType, node_field: SDocNodeField
    ) -> Markup:
        assert isinstance(node_field, SDocNodeField), node_field

        if (document_type, node_field) in self.cache:
            return self.cache[(document_type, node_field)]

        prev_part = None
        parts_output = ""
        for part in node_field.parts:
            if isinstance(part, str):
                if isinstance(prev_part, InlineLink):
                    parts_output += escape_str_after_inline_markup(part)
                else:
                    parts_output += part
            elif isinstance(part, InlineLink):
                linkable_node = (
                    self.traceability_index.get_linkable_node_by_uid(part.link)
                )
                href = self.link_renderer.render_node_link(
                    linkable_node, self.context_document, document_type
                )
                parts_output += self.fragment_writer.write_anchor_link(
                    linkable_node.get_display_title(), href
                )
            elif isinstance(part, Anchor):
                parts_output += self.template_anchor.render(
                    anchor=part,
                    traceability_index=self.traceability_index,
                    link_renderer=self.link_renderer,
                    document_type=DocumentType.DOCUMENT,
                )
            else:
                raise NotImplementedError
            prev_part = part

        output = self.fragment_writer.write(parts_output)
        self.cache[(document_type, node_field)] = output

        return output



from typing import Union

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.core.transforms.validation_error import (
    MultipleValidationErrorAsList,
)
from strictdoc.helpers.cast import assert_cast


class DeleteSectionCommand:
    def __init__(
        self,
        section: SDocSection,
        traceability_index: TraceabilityIndex,
    ) -> None:
        self.section: SDocSection = section
        self.traceability_index: TraceabilityIndex = traceability_index

    def validate(self) -> None:
        incoming_links = self.traceability_index.get_incoming_links(
            self.section
        )
        if incoming_links is not None and len(incoming_links) > 0:
            raise MultipleValidationErrorAsList(
                "NOT_RELEVANT",
                errors=[
                    "This section cannot be removed because it contains incoming links."
                ],
            )

    def perform(self) -> None:
        self.validate()

        section: SDocSection = self.section
        section_parent: Union[SDocSection, SDocDocument] = assert_cast(
            section.parent, (SDocSection, SDocDocument)
        )
        section_parent.section_contents.remove(section)

        self.traceability_index.update_last_updated()



from collections import defaultdict
from typing import Dict, List

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.core.transforms.validation_error import (
    MultipleValidationError,
)
from strictdoc.export.html.form_objects.included_document_form_object import (
    IncludedDocumentFormObject,
)


class UpdateIncludedDocumentTransform:
    def __init__(
        self,
        form_object: IncludedDocumentFormObject,
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
    ) -> None:
        self.form_object: IncludedDocumentFormObject = form_object
        self.document: SDocDocument = document
        self.traceability_index: TraceabilityIndex = traceability_index

    def perform(self) -> None:
        form_object = self.form_object
        document = self.document

        try:
            self.validate(form_object, document)
        except MultipleValidationError:
            raise

        # Update the document.
        document.title = form_object.document_title

    def validate(
        self,
        form_object: IncludedDocumentFormObject,
        document: SDocDocument,
    ) -> None:
        errors: Dict[str, List[str]] = defaultdict(list)
        assert isinstance(document, SDocDocument)
        if len(form_object.document_title) == 0:
            errors["TITLE"].append("Document title must not be empty.")

        if len(errors):
            raise MultipleValidationError(
                "Document form has not passed validation.", errors=errors
            )



import html
from typing import Any, Dict, Optional, Tuple, Union

from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    RangeMarker,
)
from strictdoc.core.source_tree import SourceFile
from strictdoc.export.html.document_type import DocumentType
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.string import create_safe_title_string


class LinkRenderer:
    def __init__(self, *, root_path: str, static_path: str) -> None:
        assert isinstance(root_path, str)
        assert isinstance(static_path, str), static_path
        assert len(static_path) > 0, static_path

        self.root_path: str = root_path
        self.static_path: str = (
            f"{root_path}/{static_path}"
            if len(root_path) > 0
            else f"{static_path}"
        )
        self.local_anchor_cache: Dict[Any, str] = {}
        self.req_link_cache: Dict[
            Tuple[DocumentType, int],
            Dict[Union[SDocDocument, SDocNode, SDocSection, Anchor], str],
        ] = {}

    def render_url(self, url: str) -> str:
        url = html.escape(url)
        if len(self.root_path) == 0:
            return url
        return self.root_path + "/" + url

    def render_static_url(self, url: str) -> str:
        return self.static_path + "/" + url

    # This rarely used helper adds slashes to the import statements within
    # <script type="module">, for example project_index/index.jinja.
    # Otherwise, scripts are not imported correctly.
    def render_static_url_with_prefix(self, url: str) -> str:
        static_url = "/" + self.static_path + "/" + url
        return static_url

    def render_local_anchor(
        self, node: Union[Anchor, SDocNode, SDocSection, SDocDocument]
    ) -> str:
        if node in self.local_anchor_cache:
            return self.local_anchor_cache[node]
        if isinstance(node, Anchor):
            return f"{self._string_to_link(node.value)}"

        local_anchor: str
        if node.reserved_uid is not None and len(node.reserved_uid) > 0:
            # UID is unique enough.
            local_anchor = self._string_to_link(node.reserved_uid)
        else:
            # If an element has no UID, provide some uniqueness.
            unique_prefix = node.context.title_number_string
            if isinstance(node, (SDocSection, SDocDocument)):
                local_anchor = (
                    f"{unique_prefix}-{self._string_to_link(node.title)}"
                )
            elif isinstance(node, SDocNode):
                if (
                    node.reserved_title is not None
                    and len(node.reserved_title) > 0
                ):
                    local_anchor = (
                        f"{unique_prefix}-"
                        f"{self._string_to_link(node.reserved_title)}"
                    )
                else:
                    local_anchor = self._string_to_link(node.reserved_mid)
            else:
                raise NotImplementedError(node)
        self.local_anchor_cache[node] = local_anchor
        return local_anchor

    def render_node_link(
        self,
        node: Union[SDocDocument, SDocNode, SDocSection, Anchor],
        context_document: Optional[SDocDocument],
        document_type: DocumentType,
        allow_local: bool = True,
    ) -> str:
        """
        Create an HTML URL for a given node.

        allow_local:     used on the DTR screen where we want to ensure that only
                         full paths are used when jumping to the DOC screen.
        """

        assert isinstance(
            node, (SDocDocument, SDocNode, SDocSection, Anchor)
        ), node

        if isinstance(node, SDocDocument):
            context_level_or_none: Optional[int] = None
            if context_document is not None:
                assert context_document.meta is not None
                context_level_or_none = context_document.meta.level

            assert node.meta is not None
            document_link = node.meta.get_html_link(
                document_type,
                context_level_or_none,
            )
            return document_link + "#_TOP"

        assert isinstance(node, (SDocNode, SDocSection, Anchor)), node
        assert isinstance(document_type, DocumentType), type(document_type)
        local_link = self.render_local_anchor(node)
        including_document = node.get_including_document()
        if (
            including_document is not None
            and including_document.is_bundle_document
        ):
            return f"#{local_link}"

        if (
            allow_local
            and context_document is not None
            and node.get_document() == context_document
        ):
            return f"#{local_link}"

        # Now two cases:
        # - Context document exists, and we want to take into account this
        # document's depth.
        # - Context document does not exist, such as when we are on a Search
        # screen. In this case, the level is simply zero.
        level: int = 0
        if context_document is not None:
            assert context_document.meta is not None
            level = context_document.meta.level

        link_cache_key = (document_type, level)
        if link_cache_key in self.req_link_cache:
            document_type_cache = self.req_link_cache[link_cache_key]
            if node in document_type_cache:
                return document_type_cache[node]
        else:
            self.req_link_cache[link_cache_key] = {}

        parent_or_including_document = node.get_parent_or_including_document()
        assert parent_or_including_document.meta is not None
        document_link = parent_or_including_document.meta.get_html_link(
            document_type,
            level,
        )
        requirement_link = f"{document_link}#{local_link}"
        self.req_link_cache[link_cache_key][node] = requirement_link
        return requirement_link

    def render_node_doxygen_link(
        self,
        node: Union[SDocDocument, SDocNode, SDocSection, Anchor],
    ) -> str:
        """
        Create a Doxygen link for a given node.

        allow_local:     used on the DTR screen where we want to ensure that only
                         full paths are used when jumping to the DOC screen.
        """

        assert isinstance(node, (SDocNode, SDocSection, Anchor)), node

        parent_or_including_document = node.get_parent_or_including_document()
        assert parent_or_including_document is not None
        assert parent_or_including_document.meta is not None

        local_link = self.render_local_anchor(node)
        document_link = parent_or_including_document.meta.get_html_doc_link()
        requirement_link = f"{document_link}#{local_link}"
        return requirement_link

    def render_requirement_link_from_source_file(
        self, node: SDocNode, source_file: SourceFile
    ) -> str:
        assert isinstance(node, SDocNode)
        assert isinstance(source_file, SourceFile)
        local_link = self.render_local_anchor(node)
        link_cache_key = (DocumentType.DOCUMENT, source_file.level)
        if link_cache_key in self.req_link_cache:
            document_type_cache = self.req_link_cache[link_cache_key]
            if node in document_type_cache:
                return document_type_cache[node]
        else:
            self.req_link_cache[link_cache_key] = {}

        document = assert_cast(node.get_document(), SDocDocument)
        assert document.meta is not None
        document_link = document.meta.get_html_link(
            DocumentType.DOCUMENT, source_file.level + 1
        )
        requirement_link = f"{document_link}#{local_link}"
        self.req_link_cache[link_cache_key][node] = requirement_link
        return requirement_link

    @staticmethod
    def render_source_file_link(
        requirement: SDocNode, requirement_source_path: str
    ) -> str:
        assert isinstance(requirement_source_path, str), requirement_source_path

        assert requirement.ng_document_reference is not None
        document: SDocDocument = assert_cast(
            requirement.ng_document_reference.get_document(), SDocDocument
        )
        assert document.meta is not None
        path_prefix = document.meta.get_root_path_prefix()
        source_file_link = (
            f"{path_prefix}/_source_files/{requirement_source_path}.html"
        )
        return source_file_link

    @staticmethod
    def render_source_file_link_from_root(requirement_source_path: str) -> str:
        assert isinstance(requirement_source_path, str)
        return f"_source_files/{requirement_source_path}.html"

    @staticmethod
    def render_source_file_link_from_root_2(source_file: SourceFile) -> str:
        assert isinstance(source_file, SourceFile)
        source_file_link = (
            "_source_files"
            "/"
            f"{source_file.in_doctree_source_file_rel_path_posix}.html"
        )
        return source_file_link

    @staticmethod
    def render_requirement_in_source_file_link(
        requirement: SDocNode,
        source_link: str,
        context_source_file: SourceFile,
    ) -> str:
        assert isinstance(source_link, str), source_link

        def get_root_path_prefix(level: int) -> str:
            assert level > 0
            return ("../" * level)[:-1]

        path_prefix = get_root_path_prefix(context_source_file.level + 1)
        source_file_link = (
            f"{path_prefix}"
            f"/_source_files"
            f"/{source_link}.html"
            f"#{requirement.reserved_uid}"
        )
        return source_file_link

    @staticmethod
    def render_requirement_in_source_file_range_link_using_id(
        req_uid: Optional[str],  # noqa: ARG004
        source_link: str,
        context_source_file: SourceFile,
        source_range: RangeMarker,
    ) -> str:
        assert isinstance(source_link, str)
        assert isinstance(context_source_file, SourceFile)
        assert len(source_link) > 0

        if req_uid is None:
            req_uid = ""

        def get_root_path_prefix(level: int) -> str:
            assert level > 0
            return ("../" * level)[:-1]

        path_prefix = get_root_path_prefix(context_source_file.level + 1)
        source_file_link = (
            f"{path_prefix}"
            f"/_source_files"
            f"/{source_link}.html"
            f"#{req_uid}"
            f"#{source_range.ng_range_line_begin}"
            f"#{source_range.ng_range_line_end}"
        )
        return source_file_link

    def render_marker_range_link(
        self,
        source_link: str,
        context_source_file: SourceFile,
        source_range: RangeMarker,
    ) -> str:
        assert isinstance(source_link, str)
        assert isinstance(context_source_file, SourceFile)
        return self.render_requirement_in_source_file_range_link_using_id(
            # No requirement UID provided because the link is formed for a range only.
            "",
            source_link,
            context_source_file,
            source_range,
        )

    def render_requirement_in_source_file_range_link(
        self,
        requirement: SDocNode,
        source_link: str,
        context_source_file: SourceFile,
        source_range: RangeMarker,
    ) -> str:
        assert isinstance(source_link, str)
        assert isinstance(context_source_file, SourceFile)
        return self.render_requirement_in_source_file_range_link_using_id(
            requirement.reserved_uid,
            source_link,
            context_source_file,
            source_range,
        )

    @staticmethod
    def _string_to_link(string: str) -> str:
        return create_safe_title_string(string)



import re
from typing import List, Union


def atoi(text: str) -> Union[int, str]:
    return int(text) if text.isdigit() else text


def alphanumeric_sort(text: str) -> List[Union[int, str]]:
    """
    Sort keys in the human (alphanumeric) order.

    alist.sort(key=alphanumeric_sort) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    """

    return [atoi(c) for c in re.split(r"(\d+)", text)]



from typing import Any, Optional, Tuple, Type, TypeVar, Union, cast, overload

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
D = TypeVar("D")
E = TypeVar("E")
F = TypeVar("F")


@overload
def assert_cast(node: Any, node_type: Type[A]) -> A: ...  # pragma: no cover


@overload
def assert_cast(  # pragma: no cover
    node: Any, node_type: Tuple[Type[B], Type[C]]
) -> Union[B, C]: ...


@overload
def assert_cast(  # pragma: no cover
    node: Any, node_type: Tuple[Type[B], Type[C], Type[D]]
) -> Union[B, C, D]: ...


@overload
def assert_cast(  # pragma: no cover
    node: Any, node_type: Tuple[Type[B], Type[C], Type[D], Type[E]]
) -> Union[B, C, D, E]: ...


@overload
def assert_cast(  # pragma: no cover
    node: Any, node_type: Tuple[Type[B], Type[C], Type[D], Type[E], Type[F]]
) -> Union[B, C, D, E, F]: ...


def assert_cast(  # pragma: no cover
    node: Any, node_type: Union[Type[Any], Tuple[Type[Any], ...]]
) -> Any:
    assert isinstance(node, node_type), (node, node_type)
    return cast(Any, node)


@overload
def assert_optional_cast(  # pragma: no cover
    node: None, node_type: Type[A]
) -> None: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: Any, node_type: Type[A]
) -> A: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: None, node_type: Tuple[Type[B], Type[C]]
) -> None: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: Any, node_type: Tuple[Type[B], Type[C]]
) -> Union[B, C]: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: None, node_type: Tuple[Type[B], Type[C], Type[D]]
) -> None: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: Any, node_type: Tuple[Type[B], Type[C], Type[D]]
) -> Union[B, C, D]: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: None, node_type: Tuple[Type[B], Type[C], Type[D], Type[E]]
) -> None: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: Any, node_type: Tuple[Type[B], Type[C], Type[D], Type[E]]
) -> Union[B, C, D, E]: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: None, node_type: Tuple[Type[B], Type[C], Type[D], Type[E], Type[F]]
) -> None: ...


@overload
def assert_optional_cast(  # pragma: no cover
    node: Any, node_type: Tuple[Type[B], Type[C], Type[D], Type[E], Type[F]]
) -> Union[B, C, D, E, F]: ...


def assert_optional_cast(
    node: Any, node_type: Union[Type[Any], Tuple[Type[Any], ...]]
) -> Optional[Any]:
    if node is not None:
        assert isinstance(node, node_type), (node, node_type)
        return cast(Any, node)
    return None



from dataclasses import dataclass
from typing import Iterator, List, Optional, Union

from markupsafe import Markup

from strictdoc import __version__
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_view import DocumentView
from strictdoc.backend.sdoc.models.model import SDocExtendedElementIF
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.document_tree_iterator import DocumentTreeIterator
from strictdoc.core.file_tree import FileOrFolderEntry
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.source_tree import SourceFile
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.html_templates import HTMLTemplates, JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.helpers.cast import assert_cast


@dataclass
class SearchScreenViewObject:
    def __init__(
        self,
        *,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
        templates: HTMLTemplates,
        search_results: List[SDocExtendedElementIF],
        search_value: str,
        error: Optional[str],
    ) -> None:
        self.traceability_index: TraceabilityIndex = traceability_index
        self.project_config: ProjectConfig = project_config
        self.templates: HTMLTemplates = templates
        self.search_results: List[SDocExtendedElementIF] = search_results
        self.search_value: str = search_value
        self.error: Optional[str] = error

        link_renderer = LinkRenderer(
            root_path="", static_path=project_config.dir_for_sdoc_assets
        )
        markup_renderer = MarkupRenderer.create(
            "RST",
            traceability_index,
            link_renderer,
            templates,
            project_config,
            None,
        )

        self.link_renderer: LinkRenderer = link_renderer
        self.markup_renderer: MarkupRenderer = markup_renderer
        self.standalone: bool = False
        self.document_tree_iterator: DocumentTreeIterator = (
            DocumentTreeIterator(
                assert_cast(traceability_index.document_tree, DocumentTree)
            )
        )
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__
        self.current_view = DocumentView.create_default(None).views[0]
        self.document_type: DocumentType = DocumentType.DOCUMENT
        self.link_document_type: DocumentType = DocumentType.DOCUMENT
        self.document: Optional[SDocDocument] = None

    def get_document_level(self) -> int:
        return 0

    def render_truncated_node_statement(self, node: SDocNode) -> Markup:
        return self.markup_renderer.render_truncated_node_statement(
            self.document_type, node
        )

    def render_screen(self, jinja_environment: JinjaEnvironment) -> Markup:
        return jinja_environment.render_template_as_markup(
            "screens/search/index.jinja", view_object=self
        )

    def is_empty_tree(self) -> bool:
        return self.document_tree_iterator.is_empty_tree()

    def iterator_files_first(self) -> Iterator[FileOrFolderEntry]:
        yield from self.document_tree_iterator.iterator_files_first()

    def render_url(self, url: str) -> Markup:
        return Markup(self.link_renderer.render_url(url))

    def render_static_url_with_prefix(self, url: str) -> str:
        return self.link_renderer.render_static_url_with_prefix(url)

    def render_node_link(
        self, incoming_link: Union[SDocDocument, SDocNode, SDocSection, Anchor]
    ) -> str:
        return self.link_renderer.render_node_link(
            incoming_link, None, DocumentType.DOCUMENT
        )

    def render_static_url(self, url: str) -> str:
        return self.link_renderer.render_static_url(url)

    def render_source_file_link_from_root_2(
        self, source_file: SourceFile
    ) -> str:
        return Markup(
            self.link_renderer.render_source_file_link_from_root_2(source_file)
        )

    def render_local_anchor(
        self, node: Union[Anchor, SDocNode, SDocSection, SDocDocument]
    ) -> str:
        return self.link_renderer.render_local_anchor(node)



import json
import os.path
from enum import Enum
from typing import Any, Dict, List, Tuple, Union

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_config import DocumentConfig
from strictdoc.backend.sdoc.models.document_from_file import DocumentFromFile
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElementField,
    RequirementFieldType,
)
from strictdoc.backend.sdoc.models.model import SDocElementIF
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.reference import (
    ChildReqReference,
    FileReference,
    ParentReqReference,
    Reference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.helpers.cast import assert_cast


class TAG(Enum):
    SECTION = 1
    REQUIREMENT = 2
    COMPOSITE_REQUIREMENT = 3


class JSONKey:
    NODES = "NODES"
    GRAMMAR = "GRAMMAR"

    OPTIONS = "_OPTIONS"


class JSONGenerator:
    def export_tree(
        self,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
        output_json_root: str,
    ) -> None:
        project_tree_dict: Dict[str, Any] = {
            "_COMMENT": (
                "Fields with _ are metadata. "
                "Fields without _ are the actual document/section/requirement/other content."
            ),
            "DOCUMENTS": [],
        }
        document_tree: DocumentTree = assert_cast(
            traceability_index.document_tree, DocumentTree
        )
        for document_ in document_tree.document_list:
            if not project_config.export_included_documents:
                if document_.document_is_included():
                    continue

            document_json_dict = self._write_document(document_)

            project_tree_dict["DOCUMENTS"].append(document_json_dict)

        path_output_json_file = os.path.join(output_json_root, "index.json")
        project_tree_json = json.dumps(project_tree_dict, indent=4)
        with open(path_output_json_file, "w") as output_json_file:
            output_json_file.write(project_tree_json + "\n")

    @classmethod
    def _write_document(cls, document: SDocDocument) -> Dict[str, Any]:
        document_dict: Dict[str, Any] = {
            "_NODE_TYPE": "DOCUMENT",
        }

        if document.mid_permanent or document.config.enable_mid:
            document_dict["MID"] = document.reserved_mid

        document_config: DocumentConfig = document.config
        if document_config:
            uid = document_config.uid
            if uid is not None:
                document_dict["UID"] = uid

            version = document_config.version
            if version:
                document_dict["VERSION"] = version

            classification = document_config.classification
            if classification is not None:
                document_dict["CLASSIFICATION"] = classification

            requirement_prefix = document_config.requirement_prefix
            if requirement_prefix is not None:
                document_dict["PREFIX"] = requirement_prefix

            root = document_config.root
            if root is not None:
                document_dict["ROOT"] = "true" if root else "false"

            enable_mid = document_config.enable_mid
            markup = document_config.markup
            auto_levels_specified = document_config.ng_auto_levels_specified
            requirement_style = document_config.requirement_style
            requirement_in_toc = document_config.requirement_in_toc
            default_view = document_config.default_view

            if (
                enable_mid is not None
                or markup is not None
                or auto_levels_specified
                or requirement_style is not None
                or requirement_in_toc is not None
                or default_view is not None
            ):
                document_dict[JSONKey.OPTIONS] = {}

                if enable_mid is not None:
                    document_dict[JSONKey.OPTIONS]["ENABLE_MID"] = (
                        True if enable_mid else False
                    )

                if markup is not None:
                    document_dict[JSONKey.OPTIONS]["MARKUP"] = markup

                if auto_levels_specified:
                    document_dict[JSONKey.OPTIONS]["AUTO_LEVELS"] = (
                        True if document_config.auto_levels else False
                    )

                if requirement_style is not None:
                    document_dict[JSONKey.OPTIONS]["VIEW_STYLE"] = (
                        requirement_style
                    )

                if requirement_in_toc is not None:
                    document_dict[JSONKey.OPTIONS]["NODE_IN_TOC"] = (
                        requirement_in_toc
                    )

                if default_view is not None:
                    document_dict[JSONKey.OPTIONS]["DEFAULT_VIEW"] = (
                        default_view
                    )

        document_dict["TITLE"] = document.title

        #
        # Grammar.
        #

        document_dict[JSONKey.GRAMMAR] = {"ELEMENTS": []}

        assert document.grammar is not None
        document_grammar: DocumentGrammar = document.grammar

        for element_ in document_grammar.elements:
            element_dict: Dict[str, Any] = {
                "NODE_TYPE": element_.tag,
                "FIELDS": [],
                "RELATIONS": [],
            }

            for grammar_field in element_.fields:
                element_dict["FIELDS"].append(
                    cls._write_grammar_field_type(grammar_field)
                )

            if len(element_.relations) > 0:
                for element_relation_ in element_.relations:
                    relation_dict: Dict[str, str] = {
                        "TYPE": element_relation_.relation_type,
                    }
                    if element_relation_.relation_role is not None:
                        relation_dict["ROLE"] = element_relation_.relation_role
                    element_dict["RELATIONS"].append(relation_dict)

            document_dict[JSONKey.GRAMMAR]["ELEMENTS"].append(element_dict)

        node_dict = JSONGenerator._write_node(document, document, ())
        document_dict[JSONKey.NODES] = node_dict[JSONKey.NODES]

        return document_dict

    @classmethod
    def _write_node(
        cls,
        node: SDocElementIF,
        document: SDocDocument,
        level_stack: Tuple[int, ...],
    ) -> Dict[str, Any]:
        if isinstance(node, SDocSection):
            section_dict: Dict[str, Any] = cls._write_section(
                node, document, level_stack
            )
            return section_dict

        elif isinstance(node, SDocNode):
            subnode_dict = cls._write_requirement(
                node=node,
                document=document,
                level_stack=level_stack,
            )
            return subnode_dict

        elif isinstance(node, SDocDocument):
            node_dict: Dict[str, Any] = cls._write_included_document(
                node,
                level_stack=level_stack,
            )

            current_number = 0
            for subnode_ in node.section_contents:
                if subnode_.ng_resolved_custom_level is None:
                    current_number += 1
                document_subnode_dict: Dict[str, Any] = cls._write_node(
                    subnode_, document, level_stack + (current_number,)
                )
                node_dict[JSONKey.NODES].append(document_subnode_dict)

            return node_dict

        elif isinstance(node, DocumentFromFile):
            resolved_document = assert_cast(
                node.resolved_document, SDocDocument
            )
            subnode_dict = cls._write_node(
                resolved_document, document, level_stack
            )
            return subnode_dict

        else:
            raise NotImplementedError

    @classmethod
    def _write_included_document(
        cls,
        node: SDocDocument,
        level_stack: Tuple[int, ...],
    ) -> Dict[str, Any]:
        node_dict: Dict[str, Any] = {
            "_TOC": cls._get_level_string(node, level_stack=level_stack),
            "_NODE_TYPE": "SECTION",
            "TITLE": node.reserved_title,
            JSONKey.NODES: [],
        }
        return node_dict

    @classmethod
    def _write_section(
        cls,
        section: SDocSection,
        document: SDocDocument,
        level_stack: Tuple[int, ...],
    ) -> Dict[str, Any]:
        assert isinstance(section, (SDocSection, SDocDocument))
        node_dict: Dict[str, Any] = {
            "_TOC": cls._get_level_string(section, level_stack),
            "_NODE_TYPE": "SECTION",
        }

        if section.mid_permanent or document.config.enable_mid:
            node_dict["MID"] = section.reserved_mid

        if section.reserved_uid is not None:
            node_dict["UID"] = section.reserved_uid

        if (
            isinstance(section, SDocSection)
            and section.custom_level is not None
        ):
            node_dict["LEVEL"] = section.custom_level

        if section.requirement_prefix is not None:
            node_dict["PREFIX"] = section.requirement_prefix

        node_dict["TITLE"] = str(section.title)
        node_dict[JSONKey.NODES] = []

        current_number = 0
        for subnode_ in section.section_contents:
            if subnode_.ng_resolved_custom_level is None:
                current_number += 1

            section_subnode_dict: Dict[str, Any] = cls._write_node(
                subnode_, document, level_stack + (current_number,)
            )
            node_dict[JSONKey.NODES].append(section_subnode_dict)

        return node_dict

    @classmethod
    def _write_requirement(
        cls,
        node: SDocNode,
        document: SDocDocument,
        level_stack: Tuple[int, ...],
    ) -> Dict[str, Any]:
        node_dict: Dict[str, Any] = {
            "_TOC": cls._get_level_string(node, level_stack),
            "_NODE_TYPE": node.node_type,
        }

        if node.mid_permanent or document.config.enable_mid:
            node_dict["MID"] = node.reserved_mid

        document_grammar = assert_cast(document.grammar, DocumentGrammar)
        element = document_grammar.elements_by_type[node.node_type]

        for element_field in element.fields:
            field_name = element_field.title
            if field_name not in node.ordered_fields_lookup:
                continue
            fields = node.ordered_fields_lookup[field_name]
            for field in fields:
                node_dict[field_name] = field.get_text_value()

        if len(node.relations) > 0:
            node_dict["RELATIONS"] = cls._write_requirement_relations(node)

        if node.is_composite:
            node_dict[JSONKey.NODES] = []

            current_number = 0
            for subnode_ in node.section_contents:
                if subnode_.ng_resolved_custom_level is None:
                    current_number += 1

                section_subnode_dict: Dict[str, Any] = cls._write_node(
                    subnode_, document, level_stack + (current_number,)
                )
                node_dict[JSONKey.NODES].append(section_subnode_dict)

        return node_dict

    @classmethod
    def _write_grammar_field_type(
        cls, grammar_field: GrammarElementField
    ) -> Dict[str, str]:
        grammar_field_dict = {
            "TITLE": grammar_field.title,
            "REQUIRED": "True" if grammar_field.required else "False",
            # FIXME: Support more grammar types.
            "TYPE": RequirementFieldType.STRING,
        }
        return grammar_field_dict

    @staticmethod
    def _write_requirement_relations(node: SDocNode) -> List[Dict[str, str]]:
        relations_list = []

        reference: Reference
        for reference in node.relations:
            relation_dict = {
                "TYPE": reference.ref_type,
            }

            if isinstance(reference, FileReference):
                ref: FileReference = reference
                file_format = ref.get_file_format()
                if file_format is not None:
                    relation_dict["FORMAT"] = file_format

                relation_dict["VALUE"] = ref.get_posix_path()

                if ref.g_file_entry.line_range is not None:
                    relation_dict["LINE_RANGE"] = (
                        str(ref.g_file_entry.line_range[0])
                        + ", "
                        + str(ref.g_file_entry.line_range[1])
                    )

            elif isinstance(reference, ParentReqReference):
                parent_reference: ParentReqReference = reference
                relation_dict["VALUE"] = parent_reference.ref_uid
                if parent_reference.role is not None:
                    relation_dict["ROLE"] = parent_reference.role

            elif isinstance(reference, ChildReqReference):
                child_reference: ChildReqReference = reference
                relation_dict["VALUE"] = child_reference.ref_uid

                if child_reference.role is not None:
                    relation_dict["ROLE"] = child_reference.role
            else:
                raise AssertionError("Must not reach here.")  # pragma: no cover

            relations_list.append(relation_dict)

        return relations_list

    @classmethod
    def _get_level_string(
        cls,
        node_: Union[SDocNode, SDocSection, SDocDocument],
        level_stack: Tuple[int, ...],
    ) -> str:
        return (
            ""
            if node_.ng_resolved_custom_level == "None"
            else (
                node_.ng_resolved_custom_level
                if node_.ng_resolved_custom_level is not None
                else ".".join(map(str, level_stack))
            )
        )



import itertools
import typing as t

T = t.TypeVar("T")


class OrderedSet(t.MutableSet[T]):
    """
    A set that preserves insertion order by internally using a dict.

    >>> OrderedSet([1, 2, "foo"])

    Source: https://github.com/bustawin/ordered-set-37.
    """

    __slots__ = ("_d",)

    def __init__(self, iterable: t.Optional[t.Iterable[T]] = None):
        self._d = dict.fromkeys(iterable) if iterable else {}

    def add(self, value: T) -> None:
        self._d[value] = None

    def clear(self) -> None:
        self._d.clear()

    def discard(self, value: T) -> None:
        self._d.pop(value, None)

    def __getitem__(self, index: int) -> T:
        try:
            return next(itertools.islice(self._d, index, index + 1))
        except StopIteration:
            raise IndexError(f"index {index} out of range") from None

    def __contains__(self, x: object) -> bool:
        return self._d.__contains__(x)

    def __len__(self) -> int:
        return self._d.__len__()

    def __iter__(self) -> t.Iterator[T]:
        return self._d.__iter__()

    def __str__(self) -> str:
        return f"{{{', '.join(str(i) for i in self)}}}"

    def __repr__(self) -> str:
        return f"<OrderedSet {self}>"

    def sort(self, key: t.Optional[t.Callable[[T], t.Any]] = None) -> None:
        key_wrapper: t.Optional[t.Callable[[t.Any], t.Any]]
        if key is not None:

            def key_wrapper(item: t.Tuple[T, None]) -> t.Any:
                return key(item[0])
        else:
            key_wrapper = None

        self._d = dict(sorted(self._d.items(), key=key_wrapper))



"""
@relation(SDOC-SRS-110, SDOC-SRS-151, scope=file)
"""

from typing import TYPE_CHECKING, List, Optional, Tuple

from strictdoc.helpers.auto_described import auto_described

if TYPE_CHECKING:
    from strictdoc.backend.sdoc.models.document import SDocDocument


@auto_described
class DocumentCustomMetadataKeyValuePair:
    def __init__(
        self,
        *,
        parent: Optional["DocumentCustomMetadata"] = None,
        key: Optional[str],
        value: Optional[str],
    ) -> None:
        _ = parent
        self.key = key
        self.value = value


@auto_described
class DocumentCustomMetadata:
    def __init__(
        self,
        *,
        parent: Optional["DocumentConfig"] = None,
        entries: List[DocumentCustomMetadataKeyValuePair],
    ) -> None:
        _ = parent
        self.entries: List[DocumentCustomMetadataKeyValuePair] = entries


@auto_described
class DocumentConfig:
    @staticmethod
    def default_config(document: Optional["SDocDocument"]) -> "DocumentConfig":
        return DocumentConfig(
            parent=document,
            version=None,
            date=None,
            uid=None,
            classification=None,
            requirement_prefix=None,
            root=None,
            enable_mid=None,
            markup=None,
            auto_levels=None,
            layout=None,
            requirement_style=None,
            requirement_in_toc=None,
            default_view=None,
            custom_metadata=None,
        )

    def __init__(
        self,
        *,
        parent: Optional["SDocDocument"],
        version: Optional[str],
        date: Optional[str],
        uid: Optional[str],
        classification: Optional[str],
        requirement_prefix: Optional[str],
        root: Optional[str],
        enable_mid: Optional[str],
        markup: Optional[str],
        auto_levels: Optional[str],
        layout: Optional[str],
        requirement_style: Optional[str],
        requirement_in_toc: Optional[str],
        default_view: Optional[str],
        custom_metadata: Optional[DocumentCustomMetadata],
        view_style_tag: Optional[str] = None,
        node_in_toc_tag: Optional[str] = None,
    ) -> None:
        self.parent = parent
        self.version: Optional[str] = version
        self.date: Optional[str] = date

        meaningful_uid: Optional[str] = None
        if uid is not None and len(uid) > 0:
            meaningful_uid = uid
        self.uid: Optional[str] = meaningful_uid

        self.classification: Optional[str] = classification
        self.requirement_prefix: Optional[str] = requirement_prefix

        self.root: Optional[bool] = (
            True if root == "True" else (False if root == "False" else None)
        )
        self.enable_mid: Optional[bool] = (
            True
            if enable_mid == "True"
            else (False if root == "False" else None)
        )

        self.markup: Optional[str] = markup
        self.auto_levels: bool = auto_levels is None or auto_levels == "On"

        if layout is not None:
            if len(layout) > 0:
                assert layout in ("Default", "Website")
            else:
                layout = None
        self.layout: Optional[str] = layout

        # Possible requirement styles:
        # Simple, Table, Rows.
        self.requirement_style: Optional[str] = requirement_style
        self.view_style_tag: Optional[str] = view_style_tag

        self.requirement_in_toc: Optional[str] = requirement_in_toc
        self.node_in_toc_tag: Optional[str] = node_in_toc_tag

        self.default_view: Optional[str] = default_view
        self.ng_auto_levels_specified = auto_levels is not None

        self.ng_line_start: int = 0
        self.ng_col_start: int = 0

        self.custom_metadata: Optional[DocumentCustomMetadata] = custom_metadata

    def get_markup(self) -> str:
        if self.markup is None:
            return "RST"
        return self.markup

    def get_requirement_style_mode(self) -> str:
        if (
            self.requirement_style is None
            or self.requirement_style == "Narrative"
        ):
            return "narrative"
        if self.requirement_style == "Plain":
            return "plain"
        if self.requirement_style in (
            "Inline",
            "Simple",
        ):
            return "simple"
        if self.requirement_style == "Table":
            return "table"
        if self.requirement_style == "Zebra":
            return "zebra"
        raise NotImplementedError(self.requirement_style)

    def is_requirement_in_toc(self) -> bool:
        return (
            self.requirement_in_toc is None or self.requirement_in_toc == "True"
        )

    def get_prefix(self) -> str:
        if self.requirement_prefix is not None:
            return self.requirement_prefix
        return "REQ-"

    def has_meta(self) -> bool:
        # TODO: When OPTIONS are not provided to a document, the self.number and
        # self.version are both None. Otherwise, they become empty strings "".
        # This issue might deserve a bug report to TextX.
        return (
            (self.uid is not None and len(self.uid) > 0)
            or (self.version is not None and len(self.version) > 0)
            or (
                self.classification is not None and len(self.classification) > 0
            )
        )

    def has_custom_metadata(self) -> bool:
        return (
            self.custom_metadata is not None
            and self.custom_metadata.entries is not None
        )

    def get_custom_metadata(self) -> List[Tuple[str, str]]:
        if (
            self.custom_metadata is not None
            and self.custom_metadata.entries is not None
        ):
            return [
                (entry.key, entry.value)
                for entry in self.custom_metadata.entries
                if entry.key is not None and entry.value is not None
            ]
        return []



from strictdoc.backend.sdoc.grammar.grammar import (
    DOCUMENT_GRAMMAR,
    FREE_TEXT_PARSER_GRAMMAR,
    SECTION_GRAMMAR,
    TEXT_TYPES_GRAMMAR,
)
from strictdoc.backend.sdoc.grammar.grammar_grammar import (
    GRAMMAR_GRAMMAR,
    GRAMMAR_WRAPPER,
)
from strictdoc.backend.sdoc.grammar.type_system import (
    STRICTDOC_BASIC_TYPE_SYSTEM,
)


class SDocGrammarBuilder:
    @staticmethod
    def _prep_grammar(grammar: str) -> str:
        return grammar.replace("'\\n'", "'\n'")

    @staticmethod
    def create_grammar() -> str:
        grammar = SDocGrammarBuilder._prep_grammar(
            DOCUMENT_GRAMMAR
            + GRAMMAR_GRAMMAR
            + SECTION_GRAMMAR
            + FREE_TEXT_PARSER_GRAMMAR
            + TEXT_TYPES_GRAMMAR
            + STRICTDOC_BASIC_TYPE_SYSTEM
        )
        return grammar

    @staticmethod
    def create_free_text_grammar() -> str:
        return SDocGrammarBuilder._prep_grammar(
            FREE_TEXT_PARSER_GRAMMAR + TEXT_TYPES_GRAMMAR
        )

    @staticmethod
    def create_grammar_grammar() -> str:
        grammar = SDocGrammarBuilder._prep_grammar(
            GRAMMAR_WRAPPER
            + GRAMMAR_GRAMMAR
            + STRICTDOC_BASIC_TYPE_SYSTEM
            + TEXT_TYPES_GRAMMAR
        )
        return grammar

    @staticmethod
    def create_raw_grammar() -> str:
        grammar = (
            DOCUMENT_GRAMMAR
            + GRAMMAR_GRAMMAR
            + SECTION_GRAMMAR
            + FREE_TEXT_PARSER_GRAMMAR
            + TEXT_TYPES_GRAMMAR
            + STRICTDOC_BASIC_TYPE_SYSTEM
        )
        return grammar



from dataclasses import dataclass
from datetime import datetime
from typing import Any, Iterator, List, Optional, Sequence, Tuple, Union

from jinja2 import Template
from markupsafe import Markup

from strictdoc import __version__
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_view import ViewElement
from strictdoc.backend.sdoc.models.grammar_element import GrammarElement
from strictdoc.backend.sdoc.models.model import SDocElementIF
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.document_iterator import DocumentIterationContext
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.document_tree_iterator import DocumentTreeIterator
from strictdoc.core.file_tree import File, FileOrFolderEntry, Folder
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.generators.view_objects.helpers import (
    screen_should_display_file,
    screen_should_display_folder,
)
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.git_client import GitClient
from strictdoc.helpers.string import interpolate_at_pattern_lazy
from strictdoc.server.helpers.turbo import render_turbo_stream


@dataclass
class DocumentScreenViewObject:
    def __init__(
        self,
        *,
        document_type: DocumentType,
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
        link_renderer: LinkRenderer,
        markup_renderer: MarkupRenderer,
        git_client: GitClient,
        standalone: bool,
    ):
        self.document_type: DocumentType = document_type
        self.link_document_type: DocumentType = DocumentType.DOCUMENT
        self.document: SDocDocument = document
        self.traceability_index: TraceabilityIndex = traceability_index
        self.project_config: ProjectConfig = project_config
        self.link_renderer: LinkRenderer = link_renderer
        self.markup_renderer: MarkupRenderer = markup_renderer
        self.git_client: GitClient = git_client
        self.standalone: bool = standalone
        self.document_iterator = self.traceability_index.get_document_iterator(
            self.document
        )
        self.document_tree_iterator: DocumentTreeIterator = (
            DocumentTreeIterator(
                assert_cast(traceability_index.document_tree, DocumentTree)
            )
        )
        self.current_view: ViewElement = document.view.get_current_view(
            project_config.view
        )
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__

        self.custom_html2pdf_template: Optional[Template] = None
        if project_config.html2pdf_template is not None:
            with open(project_config.html2pdf_template) as f_:
                self.custom_html2pdf_template = Template(f_.read())

    def has_included_document(self) -> bool:
        return len(self.document.included_documents) > 0

    def render_screen(self, jinja_environment: JinjaEnvironment) -> Markup:
        if self.document_type.is_document():
            if self.document.config.layout == "Website":
                return jinja_environment.render_template_as_markup(
                    "website/document/index.jinja", view_object=self
                )
            return jinja_environment.render_template_as_markup(
                "screens/document/document/index.jinja", view_object=self
            )
        elif self.document_type.is_table():
            return jinja_environment.render_template_as_markup(
                "screens/document/table/index.jinja", view_object=self
            )
        elif self.document_type.is_trace():
            return jinja_environment.render_template_as_markup(
                "screens/document/traceability/index.jinja", view_object=self
            )
        elif self.document_type.is_deeptrace():
            return jinja_environment.render_template_as_markup(
                "screens/document/traceability_deep/index.jinja",
                view_object=self,
            )
        elif self.document_type.is_pdf():
            return jinja_environment.render_template_as_markup(
                "screens/document/pdf/index.jinja", view_object=self
            )
        else:
            raise NotImplementedError(self.document_type)  # pragma: no cover

    def render_updated_screen(
        self, jinja_environment: JinjaEnvironment
    ) -> Markup:
        output = jinja_environment.render_template_as_markup(
            "actions/"
            "document/"
            "create_requirement/"
            "stream_created_requirement.jinja.html",
            view_object=self,
        )

        output += jinja_environment.render_template_as_markup(
            "actions/document/_shared/stream_updated_toc.jinja.html",
            view_object=self,
        )

        return output

    def render_updated_nodes_and_toc(
        self,
        nodes: Sequence[Union[SDocDocument, SDocNode]],
        jinja_environment: JinjaEnvironment,
    ) -> str:
        output: str = ""

        for node_ in nodes:
            template_folder: str
            if isinstance(node_, SDocDocument):
                template_folder = "section"
            elif isinstance(node_, SDocNode):
                if node_.is_text_node():
                    template_folder = "text_node"
                else:
                    template_folder = "node_content"
            else:
                raise NotImplementedError
            content = jinja_environment.render_template_as_markup(
                f"components/{template_folder}/index_extends_node.jinja",
                view_object=self,
                node=node_,
            )
            output += render_turbo_stream(
                content=content,
                action="replace",
                target=f"article-{node_.reserved_mid}",
            )

        toc_content = jinja_environment.render_template_as_markup(
            "screens/document/_shared/toc.jinja", view_object=self
        )
        output += render_turbo_stream(
            content=toc_content,
            action="update",
            target="frame-toc",
        )

        return output

    def render_update_document_content_with_moved_node(
        self, jinja_environment: JinjaEnvironment, moved_node: Any
    ) -> Markup:
        content = jinja_environment.render_template_as_markup(
            "screens/document/document/frame_document_content.jinja.html",
            view_object=self,
        )
        output = render_turbo_stream(
            content=content,
            action="replace",
            target="frame_document_content",
        )
        toc_content = jinja_environment.render_template_as_markup(
            "actions/document/_shared/stream_updated_toc.jinja.html",
            view_object=self,
            last_moved_node_id=moved_node.reserved_mid,
        )
        output += render_turbo_stream(
            toc_content,
            action="update",
            target="frame-toc",
        )
        return output

    def render_document_version(self) -> Optional[str]:
        if self.document.config.version is None:
            return None

        def resolver(variable_name: str) -> str:
            if variable_name == "GIT_VERSION":
                return self.git_client.get_commit_hash()
            elif variable_name == "GIT_BRANCH":
                return self.git_client.get_branch()
            return variable_name

        return interpolate_at_pattern_lazy(
            self.document.config.version, resolver
        )

    def render_document_date(self) -> Optional[str]:
        if self.document.config.date is None:
            return None

        def resolver(variable_name: str) -> str:
            if variable_name == "GIT_COMMIT_DATE":
                return self.git_client.get_commit_date()
            elif variable_name == "GIT_COMMIT_DATETIME":
                return self.git_client.get_commit_datetime()
            return variable_name

        return interpolate_at_pattern_lazy(self.document.config.date, resolver)

    def is_empty_tree(self) -> bool:
        return self.document_tree_iterator.is_empty_tree()

    def is_deeptrace(self) -> bool:
        return self.document_type.is_deeptrace()

    def has_any_nodes(self) -> bool:
        return self.document.has_any_nodes()

    def iterator_files_first(self) -> Iterator[FileOrFolderEntry]:
        yield from self.document_tree_iterator.iterator_files_first()

    def render_url(self, url: str) -> Markup:
        return Markup(self.link_renderer.render_url(url))

    def render_node_link(self, node: SDocNode) -> str:
        assert isinstance(node, SDocNode), node
        return self.link_renderer.render_node_link(
            node, self.document, self.document_type
        )

    def render_document_link(
        self,
        document: SDocDocument,
        context_document: SDocDocument,
        document_type_string: str,
    ) -> str:
        assert document is not None, document
        return self.link_renderer.render_node_link(
            document, context_document, DocumentType(document_type_string)
        )

    @staticmethod
    def render_standalone_document_link(document: SDocDocument) -> str:
        assert document.meta is not None

        root_prefix = document.meta.get_root_path_prefix()
        document_link = document.meta.get_html_standalone_document_link()
        return (
            document_link
            if len(root_prefix) == 0
            else "/".join((root_prefix, document_link))
        )

    def render_static_url(self, url: str) -> Markup:
        return Markup(self.link_renderer.render_static_url(url))

    def render_local_anchor(
        self, node: Union[Anchor, SDocNode, SDocSection, SDocDocument]
    ) -> str:
        return self.link_renderer.render_local_anchor(node)

    def render_node_statement(self, node: SDocNode) -> Markup:
        return self.markup_renderer.render_node_statement(
            self.document_type, node
        )

    def render_truncated_node_statement(self, node: SDocNode) -> Markup:
        return self.markup_renderer.render_truncated_node_statement(
            self.document_type, node
        )

    def render_node_rationale(self, node: SDocNode) -> Markup:
        return self.markup_renderer.render_node_rationale(
            self.document_type, node
        )

    def render_node_field(self, node_field: SDocNodeField) -> Markup:
        assert isinstance(node_field, SDocNodeField), node_field
        return self.markup_renderer.render_node_field(
            self.document_type, node_field
        )

    def get_page_title(self) -> str:
        return self.document_type.get_page_title()

    def get_document_level(self) -> int:
        assert self.document.meta is not None
        return self.document.meta.level

    def date_today(self) -> str:
        return datetime.today().strftime("%Y-%m-%d")

    def get_document_by_path(self, full_path: str) -> SDocDocument:
        return self.traceability_index.document_tree.get_document_by_path(
            full_path
        )

    def get_grammar_elements(self) -> List[GrammarElement]:
        assert self.document.grammar is not None
        return self.document.grammar.elements

    def table_of_contents(
        self,
    ) -> Iterator[Tuple[SDocElementIF, DocumentIterationContext]]:
        yield from self.document_iterator.table_of_contents()

    def document_has_any_toc_nodes(self) -> bool:
        return any(self.table_of_contents())

    def document_content_iterator(
        self,
    ) -> Iterator[Tuple[SDocElementIF, DocumentIterationContext]]:
        yield from self.document_iterator.all_content(
            print_fragments=True,
        )

    def should_display_folder(self, folder: Folder) -> bool:
        return screen_should_display_folder(
            folder,
            self.traceability_index,
            self.project_config,
            must_only_include_non_included_sdoc=True,
        )

    def should_display_file(self, file: File) -> bool:
        return screen_should_display_file(
            file,
            self.traceability_index,
            self.project_config,
            must_only_include_non_included_sdoc=True,
        )

    def should_display_included_documents_for_document(
        self, document: SDocDocument
    ) -> bool:
        return (
            self.project_config.export_included_documents
            and len(document.included_documents) > 0
        )

    def should_display_stable_link(
        self, node: Union[SDocDocument, SDocSection, SDocNode]
    ) -> bool:
        assert isinstance(node, (SDocDocument, SDocSection, SDocNode)), node
        return node.reserved_uid is not None

    def should_display_old_section_as_deprecated(self) -> bool:
        return self.project_config.is_new_section_behavior()

    def get_stable_link(
        self, node: Union[SDocDocument, SDocSection, SDocNode]
    ) -> str:
        """
        Get a stable link for a given node.

        An example of a link produced: ../../#SDOC_UG_CONTACT
        The copy_stable_link_button_controller.js consumes this link and
        transforms it into a link like: http://127.0.0.1:5111/?a=SDOC_UG_CONTACT.
        """

        assert isinstance(node, (SDocDocument, SDocSection, SDocNode)), node
        base_url = self.link_renderer.render_url("")
        if node.reserved_uid is not None:
            return base_url + "#" + node.reserved_uid
        if node.reserved_mid is not None and node.mid_permanent:
            return base_url + "#" + node.reserved_mid
        return base_url



"""
@relation(SDOC-SRS-150, scope=file)
"""

# mypy: disable-error-code="unreachable"

from typing import Optional

from strictdoc.backend.sdoc.models.model import (
    SDocDocumentIF,
    SDocNodeFieldIF,
    SDocNodeIF,
)
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.mid import MID


@auto_described
class Anchor:
    def __init__(
        self, parent: SDocNodeFieldIF, value: str, title: Optional[str]
    ) -> None:
        # In the grammar, the title is optional but textX passes it as an empty
        # string. Putting an assert to monitor the regressions/changes if the
        # grammar gets changed.
        assert title is not None
        # FIXME: Cannot enable this assert because the parent can also be
        #        FreeTextContainer. Refactor the types.
        # assert isinstance(parent, SDocNodeFieldIF), parent  # noqa: ERA001

        self.parent: SDocNodeFieldIF = parent
        self.value: str = value

        has_title = len(title) > 0
        self.title: str = title if has_title else value
        self.has_title = has_title

        # FIXME: Remove either mid or reserved_mid.
        self.mid: MID = MID.create()
        self.reserved_mid: MID = self.mid

    def get_display_title(
        self,
        include_toc_number: bool = True,  # noqa: ARG002
    ) -> str:
        return self.title

    def get_document(self) -> SDocDocumentIF:
        if isinstance(self.parent.parent, SDocDocumentIF):
            return self.parent.parent
        document: SDocDocumentIF = assert_cast(
            self.parent_node().get_document(), SDocDocumentIF
        )
        return document

    def get_parent_or_including_document(self) -> SDocDocumentIF:
        return self.parent_node().get_parent_or_including_document()

    def get_including_document(self) -> Optional[SDocDocumentIF]:
        return self.parent_node().get_including_document()

    def parent_node(self) -> SDocNodeIF:
        # Anchor -> SDocField -> SDocNode.
        parent_node: SDocNodeIF = assert_cast(self.parent.parent, SDocNodeIF)
        return parent_node



from typing import TYPE_CHECKING, List, Optional

from strictdoc.helpers.auto_described import auto_described

if TYPE_CHECKING:
    from strictdoc.backend.sdoc.models.document import SDocDocument


@auto_described
class ViewElementField:
    def __init__(
        self, parent: "ViewElementTags", name: str, placement: Optional[str]
    ):
        self.parent = parent
        self.name: str = name
        self.placement: Optional[str] = placement


@auto_described
class ViewElementTags:
    def __init__(
        self,
        parent: "ViewElement",
        object_type: str,
        visible_fields: List[ViewElementField],
    ):
        self.parent = parent
        self.object_type: str = object_type
        self.visible_fields: List[ViewElementField] = visible_fields


@auto_described
class ViewElementHiddenTag:
    def __init__(self, parent: "ViewElement", hidden_tag: str):
        self.parent = parent
        self.hidden_tag: str = hidden_tag


class NullViewElement:
    def includes_field(
        self,
        node_type: str,  # noqa: ARG002
        field_name: str,  # noqa: ARG002
    ) -> bool:
        return True


@auto_described()
class ViewElement(NullViewElement):
    def __init__(
        self,
        parent: "DocumentView",
        view_id: str,
        tags: List[ViewElementTags],
        hidden_tags: Optional[List[ViewElementHiddenTag]],
        name: Optional[str],
    ):
        self.parent = parent
        self.view_id: str = view_id
        self.tags: List[ViewElementTags] = tags
        self.hidden_tags: Optional[List[ViewElementHiddenTag]] = hidden_tags
        self.name: Optional[str] = name

    def includes_field(self, node_type: str, field_name: str) -> bool:
        for tag_ in self.tags:
            if tag_.object_type == node_type:
                field_: ViewElementField
                for field_ in tag_.visible_fields:
                    if field_.name == field_name:
                        return True
                break
        return False


@auto_described()
class DefaultViewElement(ViewElement):
    """
    FIXME: It is not great to provide a default implementation this way but it
           works fine for now.
    """

    def includes_field(self, node_type: str, field_name: str) -> bool:  # noqa: ARG002
        return True


@auto_described()
class DocumentView:
    def __init__(
        self, parent: Optional["SDocDocument"], views: List[ViewElement]
    ):
        self.parent = parent
        self.views: List[ViewElement] = views
        self.ng_line_start: int = 0
        self.ng_col_start: int = 0

    @staticmethod
    def create_default(parent: Optional["SDocDocument"]) -> "DocumentView":
        document_view = DocumentView(parent, [])
        document_view.views.append(
            DefaultViewElement(
                parent=document_view,
                view_id="NOT_RELEVANT",
                tags=[],
                hidden_tags=[],
                name=None,
            )
        )
        return document_view

    def get_default_view(self) -> ViewElement:
        return self.views[0]

    def get_current_view(self, view_id: Optional[str]) -> ViewElement:
        if view_id is None:
            return self.get_default_view()
        for view_element_ in self.views:
            if view_element_.view_id == view_id:
                return view_element_
        raise NotImplementedError



from dataclasses import dataclass, field
from typing import Dict, Union


@dataclass
class SourceFileStats:
    lines_total: int = 0
    lines_non_empty: int = 0
    lines_empty: int = 0
    lines_info: Dict[int, bool] = field(default_factory=dict)

    @classmethod
    def create(
        cls,
        string_or_bytes: Union[str, bytes],
    ) -> "SourceFileStats":
        if isinstance(string_or_bytes, bytes):
            string = string_or_bytes.decode("utf-8")
        else:
            string = string_or_bytes

        lines_non_empty = 0
        lines_empty = 0
        lines = string.splitlines()

        lines_info = {}
        for i_, line_ in enumerate(lines):
            line_is_not_empty = len(line_.strip()) > 0
            lines_info[i_ + 1] = line_is_not_empty
            if line_is_not_empty:
                lines_non_empty += 1
            else:
                lines_empty += 1

        return SourceFileStats(
            len(lines), lines_non_empty, lines_empty, lines_info
        )



import urllib
from dataclasses import dataclass
from typing import Optional

from markupsafe import Markup

from strictdoc import __version__
from strictdoc.core.project_config import ProjectConfig
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer


@dataclass
class DiffScreenViewObject:
    def __init__(
        self,
        *,
        project_config: ProjectConfig,
        results: bool,
        left_revision: Optional[str],
        right_revision: Optional[str],
        error_message: Optional[str],
        tab: str,
    ):
        self.project_config: ProjectConfig = project_config
        self.results: bool = results
        self.left_revision: Optional[str] = left_revision
        self.right_revision: Optional[str] = right_revision
        self.error_message: Optional[str] = error_message
        self.tab: str = tab

        link_renderer = LinkRenderer(
            root_path="", static_path=project_config.dir_for_sdoc_assets
        )
        self.link_renderer: LinkRenderer = link_renderer
        self.standalone: bool = False
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__

    def get_document_level(self) -> int:
        return 0

    @property
    def left_revision_urlencoded(self) -> str:
        return (
            urllib.parse.quote(self.left_revision)
            if self.left_revision is not None
            else ""
        )

    @property
    def right_revision_urlencoded(self) -> str:
        return (
            urllib.parse.quote(self.right_revision)
            if self.right_revision is not None
            else ""
        )

    def render_screen(self, jinja_environment: JinjaEnvironment) -> Markup:
        return jinja_environment.render_template_as_markup(
            "screens/git/index.jinja", view_object=self
        )

    def render_url(self, url: str) -> Markup:
        return Markup(self.link_renderer.render_url(url))

    def render_static_url(self, url: str) -> Markup:
        return Markup(self.link_renderer.render_static_url(url))

    def render_static_url_with_prefix(self, url: str) -> str:
        return self.link_renderer.render_static_url_with_prefix(url)



"""
@relation(SDOC-SRS-28, SDOC-SRS-2, scope=file)
"""

import datetime
import glob
import os
import sys
from typing import Any, Dict, Iterator, List, Optional, Set, Union

from textx import TextXSyntaxError

from strictdoc.backend.sdoc.error_handling import StrictDocSemanticError
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_from_file import DocumentFromFile
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.grammar_element import ReferenceType
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.model import SDocDocumentFromFileIF
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.reference import (
    ChildReqReference,
    ParentReqReference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.backend.sdoc.validations.sdoc_validator import SDocValidator
from strictdoc.backend.sdoc_source_code.caching_reader import (
    SourceFileTraceabilityCachingReader,
)
from strictdoc.core.constants import GraphLinkType
from strictdoc.core.document_finder import DocumentFinder
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.file_dependency_manager import FileDependencyManager
from strictdoc.core.file_traceability_index import FileTraceabilityIndex
from strictdoc.core.finders.source_files_finder import (
    SourceFilesFinder,
)
from strictdoc.core.graph.many_to_many_set import ManyToManySet
from strictdoc.core.graph.one_to_one_dictionary import OneToOneDictionary
from strictdoc.core.graph_database import GraphDatabase
from strictdoc.core.project_config import ProjectConfig, ProjectFeature
from strictdoc.core.query_engine.query_object import (
    QueryNullObject,
    QueryObject,
)
from strictdoc.core.query_engine.query_reader import QueryReader
from strictdoc.core.source_tree import SourceFile, SourceTree
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.core.tree_cycle_detector import TreeCycleDetector
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.deprecation_engine import DEPRECATION_ENGINE
from strictdoc.helpers.exception import StrictDocException
from strictdoc.helpers.file_modification_time import (
    get_file_modification_time,
)
from strictdoc.helpers.mid import MID
from strictdoc.helpers.parallelizer import Parallelizer
from strictdoc.helpers.timing import measure_performance, timing_decorator


class TraceabilityIndexBuilder:
    @staticmethod
    def create(
        *,
        project_config: ProjectConfig,
        parallelizer: Parallelizer,
        skip_source_files: bool = False,
    ) -> TraceabilityIndex:
        # TODO: It would be great to hide this code behind --development flag.
        # There is no need for this to be activated in the Pip-released builds.
        strict_own_files_unfiltered: Iterator[str] = glob.iglob(
            f"{project_config.get_strictdoc_root_path()}/strictdoc/**/*",
            recursive=True,
        )
        strict_own_files: List[str] = [
            f
            for f in strict_own_files_unfiltered
            if f.endswith(".html") or f.endswith(".py") or f.endswith(".jinja")
        ]
        latest_strictdoc_own_file = (
            max(strict_own_files, key=os.path.getctime)
            if len(strict_own_files) > 0
            else None
        )

        strictdoc_last_update: datetime.datetime = (
            get_file_modification_time(latest_strictdoc_own_file)
            if (latest_strictdoc_own_file is not None)
            else datetime.datetime.fromtimestamp(0)
        )
        if (
            project_config.config_last_update is not None
            and project_config.config_last_update > strictdoc_last_update
        ):
            strictdoc_last_update = project_config.config_last_update

        document_tree, asset_manager = DocumentFinder.find_sdoc_content(
            project_config=project_config, parallelizer=parallelizer
        )

        # TODO: This is rather messy, but it is better than it used to be.
        # Currently, the traceability index holds everything that is later used
        # by HTML generators:
        # - traceability index itself
        # - document tree
        # - assets
        # - runtime configuration.
        traceability_index: TraceabilityIndex = (
            TraceabilityIndexBuilder.create_from_document_tree(
                document_tree,
                project_config,
            )
        )
        traceability_index.asset_manager = asset_manager
        traceability_index.strictdoc_last_update = strictdoc_last_update

        TraceabilityIndexBuilder._filter_nodes(
            project_config=project_config, traceability_index=traceability_index
        )

        #
        # File traceability-related calculations.
        #
        if not skip_source_files and project_config.is_feature_activated(
            ProjectFeature.REQUIREMENT_TO_SOURCE_TRACEABILITY
        ):
            file_tracability_index = (
                traceability_index.get_file_traceability_index()
            )
            source_tree: SourceTree = SourceFilesFinder.find_source_files(
                project_config=project_config
            )
            source_files = source_tree.source_files
            source_file: SourceFile
            for source_file in source_files:
                with measure_performance(
                    f"Reading source: {source_file.in_doctree_source_file_rel_path}"
                ):
                    traceability_info = (
                        SourceFileTraceabilityCachingReader.read_from_file(
                            source_file.full_path, project_config
                        )
                    )

                if traceability_info:
                    traceability_index.create_traceability_info(
                        source_file,
                        traceability_info,
                    )
                    # Is file referenced by backwards links?
                    if len(traceability_info.markers) > 0:
                        source_file.is_referenced = True

            file_tracability_index.validate_and_resolve(
                traceability_index, project_config
            )

            # Iterate again to resolve if the file is referenced.
            # FIXME: Not great to iterate two times.
            for source_file in file_tracability_index.indexed_source_files():
                # Is file referenced by forward links?
                is_source_file_referenced = (
                    traceability_index.has_source_file_reqs(
                        source_file.in_doctree_source_file_rel_path_posix
                    )
                )
                if is_source_file_referenced:
                    source_file.is_referenced = True

                    source_file_reqs: Optional[List[SDocNode]] = (
                        traceability_index.get_source_file_reqs(
                            source_file.in_doctree_source_file_rel_path_posix
                        )
                    )
                    if source_file_reqs is None:
                        continue

                    for node_ in source_file_reqs:
                        node_document = assert_cast(
                            node_.get_document(), SDocDocument
                        )
                        assert node_document.meta is not None

                        traceability_index.file_dependency_manager.add_dependency(
                            source_file.full_path,
                            source_file.output_file_full_path,
                        )
                        traceability_index.file_dependency_manager.add_dependency(
                            source_file.full_path,
                            node_document.meta.output_document_full_path,
                        )
                        traceability_index.file_dependency_manager.add_dependency(
                            node_document.meta.input_doc_full_path,
                            source_file.output_file_full_path,
                        )

            traceability_index.document_tree.attach_source_tree(source_tree)

        #
        # Resolve all modification dates to support the incremental generation of
        # all artifacts.
        #

        file_dependency_manager = traceability_index.file_dependency_manager

        file_dependency_manager.resolve_modification_dates(
            traceability_index.strictdoc_last_update
        )

        return traceability_index

    @staticmethod
    @timing_decorator("Build traceability graph")
    def create_from_document_tree(
        document_tree: DocumentTree,
        project_config: ProjectConfig,
    ) -> TraceabilityIndex:
        # FIXME: Too many things going on below. Would be great to simplify this
        # workflow.
        d_01_document_iterators: Dict[
            SDocDocument, DocumentCachingIterator
        ] = {}
        d_07_file_traceability_index = FileTraceabilityIndex()

        graph_database = GraphDatabase(
            [
                (
                    GraphLinkType.MID_TO_NODE,
                    OneToOneDictionary(
                        MID,
                        (
                            SDocNode,
                            SDocSection,
                            SDocDocument,
                            InlineLink,
                            Anchor,
                        ),
                    ),
                ),
                (
                    GraphLinkType.UID_TO_NODE,
                    OneToOneDictionary(
                        str, (SDocDocument, SDocNode, SDocSection, Anchor)
                    ),
                ),
                (
                    GraphLinkType.NODE_TO_PARENT_NODES,
                    ManyToManySet(SDocNode, SDocNode),
                ),
                (
                    GraphLinkType.NODE_TO_CHILD_NODES,
                    ManyToManySet(SDocNode, SDocNode),
                ),
                (
                    GraphLinkType.NODE_TO_INCOMING_LINKS,
                    ManyToManySet(MID, InlineLink),
                ),
                (
                    GraphLinkType.DOCUMENT_TO_TAGS,
                    OneToOneDictionary(MID, dict),
                ),
            ]
        )

        file_dependency_manager: FileDependencyManager = (
            FileDependencyManager.create_from_cache(
                project_config=project_config
            )
        )

        traceability_index = TraceabilityIndex(
            document_tree,
            d_01_document_iterators,
            file_traceability_index=d_07_file_traceability_index,
            graph_database=graph_database,
            file_dependency_manager=file_dependency_manager,
        )

        found_deprecated_section: bool = False

        # It seems to be impossible to accomplish everything in just one for
        # loop. One particular problem that requires two passes: it is not
        # possible to know after one iteration which of the requirements
        # parents do not exist for each given requirement.
        #
        # Step #1:
        # - Collect a dictionary of all requirements in the document tree:
        # {req_id: req}  # noqa: ERA001
        # - Each requirement's 'parents_uids' is populated with the forward
        # declarations of its parents uids.
        # - A separate map is created: {req_id: [req_children]}
        # At this point some information is in place, but it was not known if
        # some UIDs could not be resolved which is the task of the second
        # step.
        #
        # Step #2:
        # - Check if each requirement's has valid parent relations.
        # - Resolve parent forward declarations
        # - Re-assign children declarations
        # - Detect cycles
        # - Calculate depth of both parent and child relations.
        for (
            path_to_grammar_,
            grammar_from_file_,
        ) in document_tree.map_grammars_by_filenames.items():
            try:
                SDocValidator.validate_grammar_from_file(
                    path_to_grammar_, grammar_from_file_
                )
            except StrictDocSemanticError as exc:
                print(exc.to_print_message())  # noqa: T201
                sys.exit(1)

        document: SDocDocument
        for document in document_tree.document_list:
            assert document.grammar is not None
            assert document.meta is not None

            traceability_index.file_dependency_manager.add_dependency(
                document.meta.input_doc_full_path,
                document.meta.output_document_full_path,
            )

            if document.config.view_style_tag == "REQUIREMENT_STYLE":
                DEPRECATION_ENGINE.add_message(
                    "DEPRECATED_REQUIREMENT_STYLE",
                    "WARNING: REQUIREMENT_STYLE is deprecated. Replace it to VIEW_STYLE.",
                )
            if document.config.node_in_toc_tag == "REQUIREMENT_IN_TOC":
                DEPRECATION_ENGINE.add_message(
                    "DEPRECATED_REQUIREMENT_IN_TOC",
                    "WARNING: REQUIREMENT_IN_TOC is deprecated. Replace it to NODE_IN_TOC.",
                )

            #
            # First, resolve all grammars that are imported from grammar files.
            #
            if document.grammar.import_from_file is not None:
                document_grammar: Optional[DocumentGrammar] = (
                    document_tree.get_grammar_by_filename(
                        document.grammar.import_from_file
                    )
                )
                if document_grammar is None:
                    raise StrictDocException(
                        "TraceabilityIndex: "
                        f'the document "{document.reserved_title}" '
                        "imports a grammar from a file that does not exist: "
                        f'"{document.grammar.import_from_file}". One known '
                        f"source of this error is when only a single document "
                        f"file is provided as input to the export or server "
                        f"command, rather than the containing folder. To locate "
                        f"the grammar file, StrictDoc needs to be able to "
                        f"resolve it relative to the input path."
                    )

                document.grammar.update_with_elements(document_grammar.elements)

                # This is for the backward compatibility with the existing users.
                # If the included project grammar has no TEXT element defined,
                # we add it here automatically.
                if not document.grammar.has_text_element():
                    document.grammar.add_element_first(
                        DocumentGrammar.create_default_text_element(
                            document.grammar,
                            enable_mid=document.config.enable_mid is True,
                        )
                    )

            # This is important because due to the difference between the
            # normal grammar vs imported grammar, the parent may not be set at
            # this point.
            document.grammar.parent = document

            try:
                SDocValidator.validate_document(document)
            except StrictDocSemanticError as exc:
                print(exc.to_print_message())  # noqa: T201
                sys.exit(1)

            if graph_database.has_link(
                link_type=GraphLinkType.MID_TO_NODE,
                lhs_node=document.reserved_mid,
            ):
                other_document: SDocDocument = graph_database.get_link_value(
                    link_type=GraphLinkType.MID_TO_NODE,
                    lhs_node=document.reserved_mid,
                )
                raise StrictDocException(
                    "TraceabilityIndex: "
                    "the document MID is not unique: "
                    f"{document.reserved_mid}. "
                    "All machine identifiers (MID) must be unique values. "
                    f"Affected documents:\n"
                    f"{other_document.get_debug_info()}\n"
                    f"and\n"
                    f"{document.get_debug_info()}."
                )

            graph_database.create_link(
                link_type=GraphLinkType.MID_TO_NODE,
                lhs_node=document.reserved_mid,
                rhs_node=document,
            )
            if document.uid:
                graph_database.create_link(
                    link_type=GraphLinkType.UID_TO_NODE,
                    lhs_node=document.uid,
                    rhs_node=document,
                )

            document_tags: Dict[str, int] = {}
            graph_database.create_link(
                link_type=GraphLinkType.DOCUMENT_TO_TAGS,
                lhs_node=document.reserved_mid,
                rhs_node=document_tags,
            )

            document_iterator = DocumentCachingIterator(document)
            d_01_document_iterators[document] = document_iterator

            for node, _ in document_iterator.all_content(
                print_fragments=False,
            ):
                if (
                    isinstance(node, SDocSection)
                    and not found_deprecated_section
                ):
                    found_deprecated_section = True

                    DEPRECATION_ENGINE.add_message(
                        "DEPRECATED_SECTION",
                        "WARNING: At least one document in this documentation tree "
                        "uses a deprecated [SECTION] element. "
                        "All [SECTION] elements must be renamed to [[SECTION]], "
                        "and the SECTION element must be registered in the "
                        "document grammar. "
                        "See the migration guide for more details:\n"
                        "https://strictdoc.readthedocs.io/en/latest/latest/docs/strictdoc_01_user_guide.html#SECTION-UG-NODE-MIGRATION\n"
                        "This warning will become an error in 2025 Q3.",
                    )

                if isinstance(node, SDocNode):
                    try:
                        assert document.grammar is not None
                        SDocValidator.validate_node(
                            node,
                            document_grammar=document.grammar,
                            path_to_sdoc_file=document.meta.input_doc_full_path,
                            auto_uid_mode=project_config.auto_uid_mode,
                        )
                    except StrictDocSemanticError as exc:
                        print(exc.to_print_message())  # noqa: T201
                        sys.exit(1)

                if graph_database.has_link(
                    link_type=GraphLinkType.MID_TO_NODE,
                    lhs_node=node.reserved_mid,
                ):
                    other_node: SDocDocument = graph_database.get_link_value(
                        link_type=GraphLinkType.MID_TO_NODE,
                        lhs_node=node.reserved_mid,
                    )
                    raise StrictDocException(
                        "TraceabilityIndex: "
                        "the node MID is not unique: "
                        f"{node.reserved_mid}. "
                        "All machine identifiers (MID) must be unique values. "
                        f"Affected nodes:\n"
                        f"{other_node.get_debug_info()}\n"
                        f"and\n"
                        f"{node.get_debug_info()}."
                    )
                graph_database.create_link(
                    link_type=GraphLinkType.MID_TO_NODE,
                    lhs_node=node.reserved_mid,
                    rhs_node=node,
                )

                if node.reserved_uid is not None:
                    # @relation(SDOC-SRS-29, scope=range_start)
                    if traceability_index.graph_database.has_link(
                        link_type=GraphLinkType.UID_TO_NODE,
                        lhs_node=node.reserved_uid,
                    ):
                        already_existing_node: SDocNode = (
                            traceability_index.graph_database.get_link_value(
                                link_type=GraphLinkType.UID_TO_NODE,
                                lhs_node=node.reserved_uid,
                            )
                        )
                        other_req_doc = assert_cast(
                            already_existing_node.get_document(), SDocDocument
                        )
                        if other_req_doc == document:
                            print(  # noqa: T201
                                "error: DocumentIndex: "
                                "two nodes with the same UID "
                                "exist in the same document: "
                                f'{node.reserved_uid} in "{document.title}".'
                            )
                        else:
                            print(  # noqa: T201
                                "error: DocumentIndex: "
                                "two nodes with the same UID "
                                "exist in two different documents: "
                                f'{node.reserved_uid} in "{other_req_doc.title}" '
                                f'and "{document.title}".'
                            )
                        sys.exit(1)
                    # @relation(SDOC-SRS-29, scope=range_end)

                    traceability_index.graph_database.create_link(
                        link_type=GraphLinkType.UID_TO_NODE,
                        lhs_node=node.reserved_uid,
                        rhs_node=node,
                    )

                if isinstance(node, SDocNode):
                    requirement_node: SDocNode = assert_cast(node, SDocNode)
                    if requirement_node.reserved_tags is not None:
                        for tag in requirement_node.reserved_tags:
                            document_tags.setdefault(tag, 0)
                            document_tags[tag] += 1
                    for node_field_ in node.enumerate_fields():
                        for part in node_field_.parts:
                            # The inline links are handled at the next big
                            # For loop pass because the information about
                            # all Nodes and Anchors have not been
                            # collected yet at this point.
                            # see create_inline_link below.
                            if isinstance(part, Anchor):
                                graph_database.create_link(
                                    link_type=GraphLinkType.MID_TO_NODE,
                                    lhs_node=part.mid,
                                    rhs_node=part,
                                )
                                graph_database.create_link(
                                    link_type=GraphLinkType.UID_TO_NODE,
                                    lhs_node=part.value,
                                    rhs_node=part,
                                )

        # Now iterate over the requirements again to build an in-depth map of
        # parents and children.
        requirement: SDocNode

        for document in document_tree.document_list:
            assert document.meta is not None

            document_iterator = d_01_document_iterators[document]

            for node, _ in document_iterator.all_content(
                print_fragments=False,
            ):
                if not isinstance(node, SDocNode):
                    continue

                requirement = assert_cast(node, SDocNode)

                #
                # At this point, we resolve LINKs, and the expectation is that
                # all UIDs or ANCHORS (they also have UIDs) are registered at the
                # previous pass.
                #
                for node_field_ in requirement.enumerate_fields():
                    for part in node_field_.parts:
                        if isinstance(part, InlineLink):
                            if not graph_database.has_link(
                                link_type=GraphLinkType.UID_TO_NODE,
                                lhs_node=part.link,
                            ):
                                raise StrictDocException(
                                    "DocumentIndex: "
                                    "the inline link references an "
                                    "object with an UID "
                                    "that does not exist: "
                                    f"{part.link}."
                                )
                            traceability_index.create_inline_link(part)
                if requirement.reserved_uid is None:
                    continue

                # Now it is possible to resolve parents first checking if they
                # indeed exist.
                for reference in requirement.relations:
                    if reference.ref_type == ReferenceType.FILE:
                        d_07_file_traceability_index.create_requirement_with_forward_source_links(
                            requirement
                        )
                    elif reference.ref_type == ReferenceType.PARENT:
                        parent_reference: ParentReqReference = assert_cast(
                            reference, ParentReqReference
                        )
                        parent_requirement = traceability_index.graph_database.get_link_value_weak(
                            link_type=GraphLinkType.UID_TO_NODE,
                            lhs_node=parent_reference.ref_uid,
                        )
                        if parent_requirement is None:
                            raise StrictDocException(
                                f"[DocumentIndex.create] "
                                f"Requirement {requirement.reserved_uid} "
                                f"references "
                                f"parent requirement which doesn't exist: "
                                f"{parent_reference.ref_uid}."
                            )
                        traceability_index.graph_database.create_link(
                            link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                            lhs_node=requirement,
                            rhs_node=parent_requirement,
                            edge=parent_reference.role,
                        )
                        traceability_index.graph_database.create_link(
                            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                            lhs_node=parent_requirement,
                            rhs_node=requirement,
                            edge=parent_reference.role,
                        )

                        # Set document dependencies.
                        parent_document: SDocDocument = assert_cast(
                            parent_requirement.get_document(), SDocDocument
                        )
                        if document != parent_document:
                            assert parent_document.meta is not None

                            # This is where we help the incremental generation to
                            # understand that the related documents must be
                            # re-generated together.
                            file_dependency_manager.add_dependency(
                                document.meta.input_doc_full_path,
                                parent_document.meta.output_document_full_path,
                            )
                            file_dependency_manager.add_dependency(
                                parent_document.meta.input_doc_full_path,
                                document.meta.output_document_full_path,
                            )
                    elif reference.ref_type == ReferenceType.CHILD:
                        child_reference: ChildReqReference = assert_cast(
                            reference, ChildReqReference
                        )
                        child_requirement = traceability_index.graph_database.get_link_value_weak(
                            link_type=GraphLinkType.UID_TO_NODE,
                            lhs_node=child_reference.ref_uid,
                        )
                        if child_requirement is None:
                            raise StrictDocException(
                                f"[DocumentIndex.create] "
                                f"Requirement {requirement.reserved_uid} "
                                f"references a "
                                f"child requirement that doesn't exist: "
                                f"{child_reference.ref_uid}."
                            )
                        traceability_index.graph_database.create_link(
                            link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                            lhs_node=child_requirement,
                            rhs_node=requirement,
                            edge=child_reference.role,
                        )
                        traceability_index.graph_database.create_link(
                            link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                            lhs_node=requirement,
                            rhs_node=child_requirement,
                            edge=child_reference.role,
                        )
                        # Set document dependencies.
                        child_requirement_document = assert_cast(
                            child_requirement.get_document(), SDocDocument
                        )
                        if document != child_requirement_document:
                            assert child_requirement_document.meta is not None

                            # This is where we help the incremental generation to
                            # understand that the related documents must be
                            # re-generated together.
                            file_dependency_manager.add_dependency(
                                document.meta.input_doc_full_path,
                                child_requirement_document.meta.output_document_full_path,
                            )
                            file_dependency_manager.add_dependency(
                                child_requirement_document.meta.input_doc_full_path,
                                document.meta.output_document_full_path,
                            )
                    else:
                        raise AssertionError(reference.ref_type)

        # Iterate for the third time to validate the graph against
        # requirement cycles.
        parents_cycle_detector = TreeCycleDetector()
        children_cycle_detector = TreeCycleDetector()
        for document in document_tree.document_list:
            document_iterator = d_01_document_iterators[document]

            for node, _ in document_iterator.all_content(
                print_fragments=False,
            ):
                if not isinstance(node, SDocNode):
                    continue

                requirement = assert_cast(node, SDocNode)

                if requirement.reserved_uid is None:
                    continue

                # @relation(SDOC-SRS-30, scope=range_start)
                # Detect cycles
                def parent_cycle_traverse_(node_id: str) -> Any:
                    current_node = (
                        traceability_index.graph_database.get_link_value(
                            link_type=GraphLinkType.UID_TO_NODE,
                            lhs_node=node_id,
                        )
                    )
                    return list(
                        map(
                            lambda node_: node_.reserved_uid,
                            traceability_index.graph_database.get_link_values(
                                link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                                lhs_node=current_node,
                            ),
                        )
                    )

                parents_cycle_detector.check_node(
                    requirement.reserved_uid,
                    parent_cycle_traverse_,
                )

                def child_cycle_traverse_(node_id: str) -> Any:
                    current_node = (
                        traceability_index.graph_database.get_link_value(
                            link_type=GraphLinkType.UID_TO_NODE,
                            lhs_node=node_id,
                        )
                    )
                    return list(
                        map(
                            lambda node_: node_.reserved_uid,
                            traceability_index.graph_database.get_link_values(
                                link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                                lhs_node=current_node,
                            ),
                        )
                    )

                children_cycle_detector.check_node(
                    requirement.reserved_uid,
                    child_cycle_traverse_,
                )
                # @relation(SDOC-SRS-30, scope=range_end)

        map_documents_by_input_rel_path: Dict[str, SDocDocument] = {}
        for document_ in document_tree.document_list:
            assert document_.meta is not None

            map_documents_by_input_rel_path[
                document_.meta.input_doc_full_path
            ] = document_

        # @relation(SDOC-SRS-109, scope=range_start)
        unique_document_from_file_occurences: Set[str] = set()
        for document_ in document_tree.document_list:
            document_from_file_: SDocDocumentFromFileIF
            for document_from_file_ in document_.fragments_from_files:
                traceability_index.contains_included_documents = True

                assert isinstance(document_from_file_, DocumentFromFile), (
                    document_from_file_
                )

                assert (
                    document_from_file_.resolved_full_path_to_document_file
                    is not None
                )

                if (
                    document_from_file_.resolved_full_path_to_document_file
                    not in map_documents_by_input_rel_path
                ):
                    raise StrictDocException(
                        "A document includes contains a link to another document "
                        "which is not resolved in the current documentation tree: "
                        f"'{document_from_file_.file}'. This can happen if a single "
                        f"document path is provided as input to a StrictDoc command. "
                        f"Try providing a path to a folder where all documents "
                        f"are stored."
                    )
                resolved_document: SDocDocument = (
                    map_documents_by_input_rel_path[
                        document_from_file_.resolved_full_path_to_document_file
                    ]
                )

                if (
                    document_from_file_.resolved_full_path_to_document_file
                    in unique_document_from_file_occurences
                ) and resolved_document.has_any_requirements():
                    raise StrictDocException(
                        "[DOCUMENT_FROM_FILE]: "
                        "A multiple inclusion of a document is detected. "
                        "A document that contains requirements or other nodes "
                        "can be only included once: "
                        f"{document_from_file_.file}."
                    )
                unique_document_from_file_occurences.add(
                    document_from_file_.resolved_full_path_to_document_file
                )

                document_from_file_.configure_with_resolved_document(
                    resolved_document
                )

        # @relation(SDOC-SRS-109, scope=range_end)

        return traceability_index

    @staticmethod
    def _filter_nodes(
        project_config: ProjectConfig, traceability_index: TraceabilityIndex
    ) -> None:
        if (
            project_config.filter_requirements is not None
            or project_config.filter_sections is not None
        ):
            query_reader = QueryReader()
            requirements_query_object: Union[QueryObject, QueryNullObject]
            sections_query_object: Union[QueryObject, QueryNullObject]
            try:
                if project_config.filter_requirements is not None:
                    requirements_query = query_reader.read(
                        project_config.filter_requirements
                    )
                    requirements_query_object = QueryObject(
                        requirements_query, traceability_index
                    )
                else:
                    requirements_query_object = QueryNullObject()
                if project_config.filter_sections is not None:
                    sections_query = query_reader.read(
                        project_config.filter_sections
                    )
                    sections_query_object = QueryObject(
                        sections_query, traceability_index
                    )
                else:
                    sections_query_object = QueryNullObject()
            except TextXSyntaxError:
                # FIXME: This must throw a StrictDocException.
                print("error: Cannot parse filter query.")  # noqa: T201
                sys.exit(1)
            try:
                for document in traceability_index.document_tree.document_list:
                    document_iterator = (
                        traceability_index.get_document_iterator(document)
                    )
                    for node, _ in document_iterator.all_content():
                        if isinstance(
                            node, SDocSection
                        ) and not sections_query_object.evaluate(node):
                            node.ng_whitelisted = False
                            # If the node is the last one, we check if all other
                            # nodes are filtered out and if so, mark the parent
                            # section node as not whitelisted as well.
                            if (
                                node.parent.section_contents[
                                    len(node.parent.section_contents) - 1
                                ]
                                == node
                            ):
                                if isinstance(node.parent, SDocSection):
                                    node.parent.blacklist_if_needed()

                        elif isinstance(
                            node, SDocNode
                        ) and not requirements_query_object.evaluate(node):
                            node.ng_whitelisted = False
                            # If the node is the last one, we check if all other
                            # nodes are filtered out and if so, mark the parent
                            # section node as not whitelisted as well.
                            if (
                                node.parent.section_contents[
                                    len(node.parent.section_contents) - 1
                                ]
                                == node
                            ):
                                node.parent.blacklist_if_needed()

            except (AttributeError, NameError, TypeError) as attribute_error_:
                print(  # noqa: T201
                    "error: cannot apply a filter query to a node: "
                    f"{attribute_error_}"
                )
                sys.exit(1)



from math import ceil


def round_up(number: float, decimals: int) -> float:
    assert isinstance(number, float)
    assert isinstance(decimals, int)
    assert decimals >= 0

    factor = 10**decimals

    return float(ceil(number * factor) / factor)



import ipaddress
import re

# Regex for a valid hostname (RFC 1034/1035)
HOSTNAME_REGEX = re.compile(
    r"^(?!-)[A-Za-z0-9-]{1,63}(?<!-)(\.[A-Za-z0-9-]{1,63})*$"
)


def is_valid_host(host: str) -> bool:
    """
    Validate that a string is either an IP address or a hostname.
    """
    try:
        # Check if it's a valid IP address (IPv4 or IPv6).
        ipaddress.ip_address(host)
        return True
    except ValueError:
        # If not an IP, check if it's a valid hostname.
        return bool(HOSTNAME_REGEX.fullmatch(host)) and len(host) <= 253
